
__all__ = ["Function"]


class Function:
    """This is the base class of functions that can be stored to
       the ObjectStore for later execution
    """
    def __init__(self, function=None, **kwargs):
        """Create a function from the passed function, optionally
           supplying arguments that should be bound to this function
           at call time
        """
        self._args = None
        self._func = None

        if function is not None:
            if not hasattr(function, "__call__"):
                raise TypeError("You can only create a function from "
                                "callable objects")

            self._func = function

            args = {}

            for (item, value) in kwargs.items():
                args[item] = value

            if len(args) > 0:
                self._args = args

    def __call__(self, **kwargs):
        """Call this function. Any bound arguments are added to
           'kwargs' before calling
        """
        if self._args is not None:
            for (key, value) in self._args.items():
                if key not in kwargs:
                    kwargs[key] = value

        if self._func is not None:
            return self._func(**kwargs)
        else:
            return None

    def to_data(self, args_to_data=None):
        """Return a JSON-serialisable dictionary describing this function

           Args:
                args_to_data (default=None): Arguments to pass
           Returns:
                dict: JSON serialisable dict version of this function
        """
        data = {}

        if self._func is not None:
            data["function"] = self._func.__name__
            data["module"] = self._func.__module__

            if self._args is not None:
                if args_to_data is None:
                    data["args"] = self._args
                else:
                    data["args"] = args_to_data(self._args)

        return data

    @staticmethod
    def from_data(data, args_from_data=None):
        """Return the function from the JSON-deserialised dictionary

           Args:
                data (dict): Dict to create function from
                args_to_data (default=None): Arguments to pass
           Returns:
                function: Function created from data
        """

        f = Function()

        if data is not None and len(data) > 0:
            import importlib as _importlib
            m = _importlib.import_module(data["module"])
            f._func = getattr(m, data["function"])

            if "args" in data:
                if args_from_data is None:
                    f._args = data["args"]
                else:
                    f._args = args_from_data(data["args"])

        return f




__all__ = ["Balance"]


class Balance:
    """Very simple class that holds the balance, liability and
       recievable values for an account at a point in time
    """
    def __init__(self, balance=None, liability=None, receivable=None,
                 _is_safe=False):
        """Construct, optionally specifying the starting balance,
           liability and receivable. These initialise to 0 if
           not set
        """
        if _is_safe:
            self._balance = balance
            self._liability = liability
            self._receivable = receivable
        else:
            from Acquire.ObjectStore import string_to_decimal \
                as _string_to_decimal

            self._balance = _string_to_decimal(balance, default=0)
            self._liability = _string_to_decimal(liability, default=0)
            self._receivable = _string_to_decimal(receivable, default=0)

    def balance(self):
        """Return the balance"""
        return self._balance

    def liability(self):
        """Return the liability"""
        return self._liability

    def receivable(self):
        """Return the receivable"""
        return self._receivable

    def available(self, overdraft_limit=None):
        """Return the available balance (balance - liability)"""
        if overdraft_limit is None:
            return self.balance() - self.liability()
        else:
            from Acquire.Accounting import create_decimal as _create_decimal
            overdraft_limit = _create_decimal(overdraft_limit)
            return self.balance() - self.liability() + overdraft_limit

    def is_overdrawn(self, overdraft_limit=None):
        """Return whether or not this balance is overdrawn"""
        return self.available(overdraft_limit=overdraft_limit) < 0

    def __str__(self):
        return "Balance(balance=%s, liability=%s, receivable=%s)" % \
                (self.balance(), self.liability(), self.receivable())

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """Comparison"""
        if type(other) is Balance:
            return self._balance == other._balance and \
                   self._liability == other._liability and \
                   self._receivable == other._receivable
        else:
            return False

    def __add__(self, other):
        """Add balances together"""
        if type(other) is Balance:
            return Balance(balance=self._balance+other._balance,
                           liability=self._liability+other._liability,
                           receivable=self._receivable+other._receivable,
                           _is_safe=True)

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        if type(other) is _TransactionInfo:
            balance = self._balance
            liability = self._liability
            receivable = self._receivable

            if other.is_credit():
                balance += other.value()
            elif other.is_debit():
                balance -= other.value()
            elif other.is_liability():
                liability += other.value()
            elif other.is_accounts_receivable():
                receivable += other.value()
            elif other.is_received_receipt():
                balance -= other.receipted_value()
                liability -= other.original_value()
            elif other.is_sent_receipt():
                balance += other.receipted_value()
                receivable -= other.original_value()
            elif other.is_received_refund():
                balance += other.value()
            elif other.is_sent_refund():
                balance -= other.value()

            return Balance(balance=balance, liability=liability,
                           receivable=receivable, _is_safe=True)

        from Acquire.Accounting import Transaction as _Transaction
        if type(other) is _Transaction:
            return Balance(balance=self._balance+other.value(),
                           liability=self._liability,
                           receivable=self._receivable,
                           _is_safe=True)

        from Acquire.Accounting import create_decimal as _create_decimal
        value = _create_decimal(other)
        return Balance(balance=self._balance+value,
                       liability=self._liability,
                       receivable=self._receivable,
                       _is_safe=True)

    def __sub__(self, other):
        """Add balances together"""
        if type(other) is type(Balance):
            return Balance(balance=self._balance-other._balance,
                           liability=self._liability-other._liability,
                           receivable=self._receivable-other._receivable,
                           _is_safe=True)

        from Acquire.Accounting import Transaction as _Transaction
        if type(other) is _Transaction:
            return Balance(balance=self._balance-other.value(),
                           liability=self._liability,
                           receivable=self._receivable,
                           _is_safe=True)

        from Acquire.Accounting import create_decimal as _create_decimal
        value = _create_decimal(other)
        return Balance(balance=self._balance+value,
                       liability=self._liability,
                       receivable=self._receivable,
                       _is_safe=True)

    @staticmethod
    def total(balances):
        """Return the sum of the passed balances"""
        from Acquire.Accounting import create_decimal as _create_decimal
        balance = _create_decimal(0)
        liability = _create_decimal(0)
        receivable = _create_decimal(0)

        for b in balances:
            if type(b) is not Balance:
                raise TypeError("You can only sum Balance objects!")

            balance += b._balance
            liability += b._liability
            receivable += b._receivable

        return Balance(balance=balance, liability=liability,
                       receivable=receivable, _is_safe=True)

    def to_data(self):
        """Return this balance as a JSON-serialisable object"""
        data = {}

        from Acquire.ObjectStore import decimal_to_string \
            as _decimal_to_string
        data["balance"] = _decimal_to_string(self._balance)
        data["liability"] = _decimal_to_string(self._liability)
        data["receivable"] = _decimal_to_string(self._receivable)

        return data

    @staticmethod
    def from_data(data):
        """Construct a balance from the passed json-deserialised object"""
        if data is None or len(data) == 0:
            return Balance()

        from Acquire.ObjectStore import string_to_decimal \
            as _string_to_decimal

        balance = _string_to_decimal(data["balance"])
        liability = _string_to_decimal(data["liability"])
        receivable = _string_to_decimal(data["receivable"])

        return Balance(balance=balance, liability=liability,
                       receivable=receivable, _is_safe=True)




__all__ = ["Account"]


def _account_root():
    return "accounting/accounts"


def _get_last_day(datetime):
    """Return the start of the day before 'datetime', e.g.
       _get_last_day(April 1st) will return March 31st
    """
    import datetime as _datetime
    datetime = datetime - _datetime.timedelta(days=1)
    return _datetime.datetime(year=datetime.year, month=datetime.month,
                              day=datetime.day, tzinfo=datetime.tzinfo)


def _get_last_month(datetime):
    """Return the date at the start of the month before 'datetime', e.g.
       _get_last_month(March 21st) will return February 1st
    """
    import datetime as _datetime
    datetime = datetime.replace(day=1) - _datetime.timedelta(days=1)
    return _datetime.datetime(year=datetime.year, month=datetime.month,
                              day=1, tzinfo=datetime.tzinfo)


def _get_hourly_datetime(datetime):
    """Return the datetime for the top of the hour of 'datetime',
       e.g. 5.42pm would return 5.00pm
    """
    from Acquire.ObjectStore import datetime_to_datetime \
        as _datetime_to_datetime
    import datetime as _datetime
    datetime = _datetime_to_datetime(datetime)
    return _datetime.datetime(year=datetime.year,
                              month=datetime.month,
                              day=datetime.day,
                              hour=datetime.hour,
                              tzinfo=datetime.tzinfo)


def _get_key_from_hour(start, datetime):
    """Return a key encoding the passed date, starting the key with 'start',
       but only up unto the specified hour
    """
    from Acquire.ObjectStore import datetime_to_datetime \
        as _datetime_to_datetime
    datetime = _datetime_to_datetime(datetime)
    return "%s/%sT%02d" % (start, datetime.date().isoformat(), datetime.hour)


def _get_key_from_day(start, datetime):
    """Return a key encoding the passed date, starting the key with 'start',
       but only up until the specified day
    """
    from Acquire.ObjectStore import datetime_to_datetime \
        as _datetime_to_datetime
    datetime = _datetime_to_datetime(datetime)
    return "%s/%4d-%02d-%02d" % (start, datetime.year, datetime.month,
                                 datetime.day)


def _get_key_from_month(start, datetime):
    """Return a key encoding the passed date, starting the key with 'start',
       but only up to the specified month
    """
    from Acquire.ObjectStore import datetime_to_datetime \
        as _datetime_to_datetime
    datetime = _datetime_to_datetime(datetime)
    return "%s/%4d-%02d" % (start, datetime.year, datetime.month)


def _get_hour_from_key(key):
    """Return the date that is encoded in the passed key"""
    import re as _re
    m = _re.search(r"(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d)", key)

    if m:
        from Acquire.ObjectStore import date_and_time_to_datetime \
            as _date_and_time_to_datetime
        import datetime as _datetime

        return _date_and_time_to_datetime(
                    _datetime.date(year=int(m.groups()[0]),
                                   month=int(m.groups()[1]),
                                   day=int(m.groups()[2])),
                    _datetime.time(hour=int(m.groups()[3])))
    else:
        from Acquire.Accounting import AccountError
        raise AccountError("Could not find a date in the key '%s'" % key)


def _get_datetime_from_key(key):
    """Return the datetime that is encoded in the passed key

       Args:
            key(st: obj: `str`): Key to search for datetime
       Returns:
            datetime: detailed datetime object read from key
    """
    import re as _re
    m = _re.search(r"(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)\.(\d+)",
                   key)

    if m:
        from Acquire.ObjectStore import date_and_time_to_datetime \
            as _datetime_to_datetime
        import datetime as _datetime

        return _datetime_to_datetime(
                    _datetime.datetime(year=int(m.groups()[0]),
                                       month=int(m.groups()[1]),
                                       day=int(m.groups()[2]),
                                       hour=int(m.groups()[3]),
                                       minute=int(m.groups()[4]),
                                       second=int(m.groups()[5]),
                                       microsecond=int(m.groups()[6])))
    else:
        from Acquire.Accounting import AccountError
        raise AccountError("Could not find a datetime in the key '%s'" % key)


def _sum_transactions(transactions):
    """Internal function that sums all of the transactions identified
    by the passed keys.  by the passed keys. This returns a tuple of
    (balance, liability, receivable, spent)

        Args:
            keys (:obj:`list`): List of keys to parse
        Returns:
            tuple (:obj:`Decimal`, Decimal, Decimal, Decimal): balance,
            liability, receivable, spent_today

    """
    from Acquire.Accounting import Balance as _Balance
    from Acquire.Accounting import TransactionInfo as _TransactionInfo

    balance = _Balance()

    for transaction in transactions:
        if not isinstance(transaction, _TransactionInfo):
            transaction = _TransactionInfo(transaction)

        balance = balance + transaction

    return balance


class Account:
    """This class represents a single account in the ledger. It has a balance,
       and a record of the set of transactions that have been applied.

       The account really holds two accounts: the liability account and
       actual capital account. We combine both together into a single
       account to ensure that updates occur atomically

       All data for this account is stored in the object store

       The account has a set of ACLRules that specify who can
       read and write to the account (writing implies has spend
       authority), and who owns the account (can change ACLRules)
    """
    def __init__(self, name=None, description=None, uid=None,
                 aclrules=None, group_name=None, bucket=None):
        """Construct the account. If 'uid' is specified, then load the account
           from the object store (so 'name' and 'description' should be "None")
           You can also supply the ACLRules that will be used to control
           access to this account. If these are not specified then
           ACLRules.inherit() will be used, with rules inherited from the
           Accounts group that contains this Account

            Args:
                name (:obj:`str`, default=None): Name on the account
                description (:obj:`str`, default=None): Description of account
                uid (UID): Unique ID for account, if used do not pass name or
                description
                bucket (dict): contains data for bucket

            Returns:
                None

        """
        self._name = None
        self._description = None
        self._last_update = {}
        self._uid = None
        self._group_name = None

        if uid is not None:
            self._uid = str(uid)
            bucket = self._get_account_bucket(bucket)
            self._load_account(bucket)

            if name:
                if name != self.name():
                    from Acquire.Accounting import AccountError
                    raise AccountError(
                        "This account name '%s' does not match what you "
                        "expect! '%s'" % (self.name(), name))

            if description:
                if description != self.description():
                    from Acquire.Accounting import AccountError
                    raise AccountError(
                        "This account description '%s' does not match what "
                        "you expect! '%s'" % (self.description(), description))

        elif name is not None:
            self._uid = None
            self._create_account(name=name, description=description,
                                 group_name=group_name, aclrules=aclrules,
                                 bucket=bucket)

    def is_null(self):
        """Return whether or not this is a null account"""
        return self._uid is None

    def _get_now(self, now=None):
        """Return the time of 'now' (or actual now if not passed)"""
        if now is None:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            return _get_datetime_now()
        else:
            from Acquire.ObjectStore import datetime_to_datetime \
                as _datetime_to_datetime
            return _datetime_to_datetime(now)

    def _get_account_bucket(self, bucket=None):
        """Return the bucket into which to write this account,
           or 'bucket' if it is not None
        """
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            return _get_service_account_bucket()
        else:
            return bucket

    def __str__(self):
        if self._uid is None:
            return "Account::null"
        else:
            return "Account(%s|%s|%s)" % (self._name, self._description,
                                          self._uid)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _create_account(self, name, description, group_name, aclrules,
                        bucket=None):
        """Create the account from scratch"""
        if name is None or description is None:
            from Acquire.Accounting import AccountError
            raise AccountError(
                "You must pass both a name and description to create a new "
                "account")

        if self._uid is not None:
            from Acquire.Accounting import AccountError
            raise AccountError("You cannot create an account twice!")

        from Acquire.Identity import ACLRules as _ACLRules
        if aclrules is None:
            aclrules = _ACLRules.inherit()
        elif not isinstance(aclrules, _ACLRules):
            raise TypeError("The aclrules must be type ACLRules")

        from Acquire.Accounting import create_decimal as _create_decimal
        from Acquire.ObjectStore import create_uuid as _create_uuid

        bucket = self._get_account_bucket(bucket)

        self._uid = _create_uuid()
        self._name = str(name)
        self._description = str(description)
        self._overdraft_limit = _create_decimal(0)
        self._maximum_daily_limit = 0
        self._aclrules = aclrules

        if group_name is None:
            self._group_name = None
        else:
            self._group_name = str(group_name)

        # make sure that this is saved to the object store
        self._save_account(bucket)

    def _get_transactions_between(self, start_datetime, end_datetime,
                                  bucket=None):
        """Return all of the object store keys for transactions in this
           account beteen 'start_datetime' and 'end_datetime' (inclusive, e.g.
           start_datetime < transaction <= end_datetime). This will return an
           empty list if there were no transactions in this time
        """
        # convert both times to UTC
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        import datetime as _datetime

        if start_datetime is None or end_datetime is None:
            raise ValueError("NULL %s | %s" % (start_datetime, end_datetime))

        start_datetime = _datetime_to_datetime(start_datetime)
        end_datetime = _datetime_to_datetime(end_datetime)

        # get the day for each time
        start_day = start_datetime.toordinal()
        end_day = end_datetime.toordinal()

        if end_datetime.time() == _datetime.time():
            # this ends on midnight of the first day - do not
            # include this last day as nothing will match
            end_day -= 1

        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import date_to_string as _date_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import TransactionInfo as _TransactionInfo

        bucket = self._get_account_bucket()

        num_days = end_day - start_day

        if num_days < 7:
            # sufficiently few days that a day-by-day search is enough
            transactions = []
            for day in range(start_day, end_day+1):
                day_date = _datetime.datetime.fromordinal(day)
                day_string = _date_to_string(day_date)

                prefix = "%s/%s" % (self._transactions_key(), day_string)

                try:
                    keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                             prefix=prefix)
                except:
                    keys = []

                for key in keys:
                    transaction = _TransactionInfo.from_key(key)
                    datetime = transaction.datetime()
                    if datetime > start_datetime and datetime <= end_datetime:
                        transactions.append(transaction)

            return transactions

        # elif num_days < 300:  Try a better algorithm for weeks and months

        else:
            # likely more than years - easier to just scan all transactions
            # on the account
            prefix = self._transactions_key()

            try:
                keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                         prefix=prefix)
            except:
                keys = []

            transactions = []

            for key in keys:
                transaction = _TransactionInfo.from_key(key)
                datetime = transaction.datetime()
                if datetime > start_datetime and datetime <= end_datetime:
                    transactions.append(transaction)

            return transactions

    def _get_balance_key(self, now=None):
        """Return the balance key for the passed time. This is the key
           into the object store of the object that holds the starting
           balance for the account on the hour of the passed datetime.
           If 'now' is None, then the key for actual now is returned
        """

        if self.is_null():
            return None
        else:
            return _get_key_from_hour(start=self._balance_key(),
                                      datetime=self._get_now(now))

    def _find_last_balance_key(self, now=None, bucket=None):
        """Return the key containing the last hourly balance update before
           'now' (defaults to actual now if not set)
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        now = self._get_now(now)
        bucket = self._get_account_bucket(bucket)
        start = self._balance_key()

        # look for any balance keys from today
        prefix = _get_key_from_day(start=start, datetime=now)
        try:
            keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                     prefix=prefix)
        except:
            keys = []

        if len(keys) > 0:
            keys.sort()
            return keys[-1]

        # look for any balance keys from yesterday
        # (we do this to stop big lookups for active accounts when
        #  we jump between months)
        now = _get_last_day(now)
        prefix = _get_key_from_day(start=start, datetime=now)
        try:
            keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                     prefix=prefix)
        except:
            keys = []

        if len(keys) > 0:
            keys.sort()
            return keys[-1]

        # look for any balance keys from this month
        prefix = _get_key_from_month(start=start, datetime=now)
        try:
            keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                     prefix=prefix)
        except:
            keys = []

        if len(keys) > 0:
            keys.sort()
            return keys[-1]

        for _ in range(0, 6):
            now = _get_last_month(now)
            prefix = _get_key_from_month(start=start, datetime=now)
            try:
                keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                         prefix=prefix)
            except:
                keys = []

            if len(keys) > 0:
                keys.sort()
                return keys[-1]

        # wow - no balance keys at all over the last 6 months. Look for
        # *any* balance keys
        try:
            keys = _ObjectStore.get_all_object_names(bucket=bucket,
                                                     prefix=prefix)
        except:
            keys = []

        if len(keys) > 0:
            keys.sort()
            # can only return the latest key before 'now'
            for i in range(len(keys)-1, 0, -1):
                key = keys[i]
                hourly_time = _get_hour_from_key(key)

                if hourly_time < now:
                    return key

        # no balance keys at all! Set a balance key for the beginning of time
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        from Acquire.Accounting import Balance as _Balance
        import datetime as _datetime
        hourly_time = _datetime_to_datetime(_datetime.datetime.fromordinal(1))

        hourly_key = self._get_balance_key(now=hourly_time)
        hourly_balance = _Balance()
        _ObjectStore.set_object_from_json(bucket=bucket, key=hourly_key,
                                          data=hourly_balance.to_data())

        return hourly_key

    def _get_hourly_balance(self, now=None, bucket=None):
        """Calculate and return the balance at the top of the hour
           for 'now' (defaults to actually now if not specified)
        """
        now = self._get_now(now)
        hourly_key = self._get_balance_key(now)

        if hourly_key in self._last_update:
            return self._last_update[hourly_key]["hourly_balance"]

        from Acquire.Accounting import Balance as _Balance
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        bucket = self._get_account_bucket(bucket)

        try:
            data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                     key=hourly_key)
            hourly_balance = _Balance.from_data(data)
        except:
            hourly_balance = None

        hourly_now_time = _get_hourly_datetime(now)

        if hourly_balance is None:
            # look for the last balance key...
            last_balance_key = self._find_last_balance_key(now=now,
                                                           bucket=bucket)

            if last_balance_key is None:
                from Acquire.Accounting import AccountError
                raise AccountError(
                    "The first balance of the account %s has not been set?"
                    % str(self))

            data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                     key=last_balance_key)

            last_balance = _Balance.from_data(data)
            last_balance_time = _get_hour_from_key(last_balance_key)

            transactions = self._get_transactions_between(
                                        start_datetime=last_balance_time,
                                        end_datetime=hourly_now_time)

            total = _sum_transactions(transactions)

            hourly_balance = last_balance + total

        _ObjectStore.set_object_from_json(bucket=bucket,
                                          key=hourly_key,
                                          data=hourly_balance.to_data())

        self._last_update[hourly_key] = \
            {"hourly_balance": hourly_balance,
             "last_update_time": hourly_now_time,
             "last_update_balance": hourly_balance}

        return hourly_balance

    def balance(self, now=None, bucket=None):
        """Get the balance of the account at 'now' (defaults to actually now).
           This returns a Balance object for the balance, that includes
           (1) the current real balance of the account,
           neglecting any outstanding liabilities or accounts receivable,
           (2) the current total liabilities,
           and (3) the current total accounts receivable

           where 'liability' is the current total liabilities,
           where 'receivable' is the current total accounts receivable, and
           where 'spent_today' is how much has been spent today (from midnight
           until now)

           Args:
                bucket (dict, default=None): Bucket to use for calculations

            Returns:
                tuple (Decimal, Decimal, Decimal, Decimal): balance, liability,
                receivable, spent_today
        """
        now = self._get_now(now)
        bucket = self._get_account_bucket(bucket)

        # get the key to the hourly balance for now
        hourly_key = self._get_balance_key(now)

        try:
            hourly_update = self._last_update[hourly_key]
        except:
            hourly_update = None

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        bucket = self._get_account_bucket()

        if hourly_update is None:
            hourly_balance = self._get_hourly_balance(bucket=bucket,
                                                      now=now)
            last_update_time = _get_hourly_datetime(now)
            last_update_balance = hourly_balance
        else:
            hourly_balance = hourly_update["hourly_balance"]
            last_update_time = hourly_update["last_update_time"]
            last_update_balance = hourly_update["last_update_balance"]

        if last_update_time >= now:
            # the last update of this balance was in the future - go from
            # the current hour
            last_update_time = _get_hourly_datetime(now)
            last_update_balance = hourly_balance

        # next, get the transactions that have taken place since the last
        # update and sum them to get the current balance
        transactions = self._get_transactions_between(
                                 start_datetime=last_update_time,
                                 end_datetime=now, bucket=bucket)

        total = last_update_balance + _sum_transactions(transactions)

        self._last_update[hourly_key] = {"hourly_balance": hourly_balance,
                                         "last_update_time": now,
                                         "last_update_balance": total}

        return total

    def name(self):
        """Return the name of this account

            Returns:
                str or None: Name of account if account not null, else None


        """
        if self.is_null():
            return None

        return self._name

    def group_name(self):
        """Return the name of the Accounts group in which this
           account belongs. An Account can only exist in a single
           Accounts Group at a time
        """
        if self.is_null():
            return None

        return self._group_name

    def description(self):
        """Return the description of this account

            Returns:
                str or None: Description of account if account not null,
                else None
        """
        if self.is_null():
            return None

        return self._description

    def uid(self):
        """Return the UID for this account.

            Returns:
                str: UID

        """
        return self._uid

    def assert_valid_authorisation(self, authorisation, resource=None,
                                   accept_partial_match=False):
        """Assert that the passed authorisation is valid for this
           account

            Args:
                authorisation (Authorisation): authorisation
                object to be used for account

            Returns:
                None
        """
        if authorisation is None:
            raise PermissionError("You need to supply a valid authorisation!")

        from Acquire.Identity import Authorisation as _Authorisation
        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The passed authorisation must be an "
                            "Authorisation")

        user_guid = None

        identifiers = None

        if not authorisation.is_null():
            identifiers = authorisation.verify(
                                 resource=resource,
                                 accept_partial_match=accept_partial_match,
                                 return_identifiers=True)
            user_guid = authorisation.user_guid()

        upstream = None

        if self.group_name() is not None:
            from Acquire.Accounting import Accounts as _Accounts
            group = _Accounts(user_guid=user_guid, group=self.group_name())
            upstream = group.aclrules().resolve(must_resolve=False,
                                                identifiers=identifiers)

        aclrule = self._aclrules.resolve(must_resolve=True,
                                         upstream=upstream,
                                         identifiers=identifiers)

        if not aclrule.is_writeable():
            raise PermissionError(
                "You do not have permission to write (draw funds) from "
                "this account!")

    def _get_safe_now(self):
        """This function returns the current time. It avoids dangerous
           times (when the system may be updating) by sleeping through
           those times (i.e. it will sleep from HH:59:58 until HH+1:00:01)
        """
        now = self._get_now()

        # don't allow any transactions in the last 2 seconds of the hour, as
        # we will sum up the day balance at the top of each hour, and
        # don't want to risk any late transactions from messing up the
        # accounting. We also don't want any transactions at exactly the
        # top of the hour in case they get missed out when getting
        # transactions with date ranges
        while (now.minute == 59 and now.second >= 58) or \
              (now.minute == 0 and now.second == 0 and now.microsecond < 10):
            import time as _time
            # sleep in quarter-second increments to minimise disruption
            _time.sleep(0.25)
            new_now = self._get_now()

            if new_now == now:
                # now time has passed - we are being tested, so let's
                # pretend that time has gone forwards
                import datetime as _datetime
                new_now = now + _datetime.timedelta(seconds=1)

            now = new_now

        return now

    def _credit_refund(self, debit_note, refund, bucket=None):
        """Credit the value of the passed 'refund' to this account. The
           refund must be for a previous completed debit, hence the
           original debitted value is returned to the account.

            Args:
                debit_note (DebitNote): Note to be used for
                refund
                refund (Refund): Refund holding value to be refunded
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime): Return the UID and current time
        """
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(refund, _Refund):
            raise TypeError("The passed refund must be a Refund")

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if refund.is_null():
            return

        if refund.value() != debit_note.value():
            raise ValueError("The refunded value does not match the value "
                             "of the debit note: %s versus %s" %
                             (refund.value(), debit_note.value()))

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                        _TransactionCode.RECEIVED_REFUND,
                                        refund.value())

        # create a UID and datetime for this credit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this credit later. The key is made
        # up from the iso format of the datetime of the credit
        # and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem
        from Acquire.ObjectStore import create_uuid as _create_uuid

        datetime_key = _datetime_to_string(now)
        uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

        item_key = "%s/%s/%s" % (self._transactions_key(), uid, encoded_value)
        l = _LineItem(debit_note.uid(), refund.authorisation())

        bucket = self._get_account_bucket()
        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit_refund(self, refund, bucket=None):
        """Debit the value of the passed 'refund' from this account. The
           refund must be for a previous completed credit. There is a risk
           that this value has been spent, so this is one of the only
           functions that allows a balance to drop below an overdraft or
           other limit (as the refund should always succeed).

            Args:
                refund (Refund): Refund note to be processed
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime): UID and current time
        """
        from Acquire.Accounting import Refund as _Refund

        if not isinstance(refund, _Refund):
            raise TypeError("The passed refund must be a Refund")

        if refund.is_null():
            return

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(_TransactionCode.SENT_REFUND,
                                                refund.value())

        bucket = self._get_account_bucket()

        while True:
            # create a UID and datetime for this debit and record
            # it in the account
            now = self._get_safe_now()

            # and to create a key to find this debit later. The key is made
            # up from the date and  of the debit and a random string
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Accounting import LineItem as _LineItem
            from Acquire.ObjectStore import create_uuid as _create_uuid

            datetime_key = _datetime_to_string(now)
            uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

            item_key = "%s/%s/%s" % (self._transactions_key(),
                                     uid, encoded_value)
            l = _LineItem(uid, refund.authorisation())

            now2 = self._get_safe_now()

            if now2.hour == now.hour:
                # we have not moved into the next hour
                break

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _credit_receipt(self, debit_note, receipt, bucket=None):
        """Credit the value of the passed 'receipt' to this account. The
           receipt must be for a previous provisional credit, hence the
           money is awaiting transfer from accounts receivable.

            Args:
                debit_note (DebitNote): Holds the value of the credit
                to be applied to the account, value must match that of receipt
                receipt (Receipt): Receipt holding the value of the credit
                that is to be applied to account
                TODO - improve bucket docs
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime): UID and current time
        """
        from Acquire.Accounting import Receipt as _Receipt
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(receipt, _Receipt):
            raise TypeError("The passed receipt must be a Receipt")

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if receipt.is_null():
            return

        if receipt.receipted_value() != debit_note.value():
            raise ValueError("The receipted value does not match the value "
                             "of the debit note: %s versus %s" %
                             (receipt.receipted_value(), debit_note.value()))

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.SENT_RECEIPT,
                                    receipt.value(), receipt.receipted_value())

        bucket = self._get_account_bucket()

        while True:
            # create a UID and datetime for this credit and record
            # it in the account
            now = self._get_safe_now()

            # and to create a key to find this credit later. The key is made
            # up from the isoformat datetime of the credit and a random string
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Accounting import LineItem as _LineItem
            from Acquire.ObjectStore import create_uuid as _create_uuid

            datetime_key = _datetime_to_string(now)
            uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

            item_key = "%s/%s/%s" % (self._transactions_key(),
                                     uid, encoded_value)
            l = _LineItem(debit_note.uid(), receipt.authorisation())

            now2 = self._get_safe_now()

            if now2.hour == now.hour:
                # we have not moved into another hour
                break

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit_receipt(self, receipt, bucket=None):
        """Debit the value of the passed 'receipt' from this account. The
           receipt must be for a previous provisional debit, hence
           the money should be available.

            Args:
                receipt (Receipt): holds the value
                to be debited from the account
                TODO - improve bucket docs
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime): UID and current time
        """
        from Acquire.Accounting import Receipt as _Receipt

        if not isinstance(receipt, _Receipt):
            raise TypeError("The passed receipt must be a Receipt")

        if receipt.is_null():
            return

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.RECEIVED_RECEIPT,
                                    receipt.value(), receipt.receipted_value())

        bucket = self._get_account_bucket()

        # create a UID and datetime for this debit and record
        # it in the account
        while True:
            now = self._get_safe_now()

            # and to create a key to find this debit later. The key is made
            # up from the isoformat datetime of the debit and a random string
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Accounting import LineItem as _LineItem
            from Acquire.ObjectStore import create_uuid as _create_uuid

            datetime_key = _datetime_to_string(now)
            uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

            item_key = "%s/%s/%s" % (self._transactions_key(),
                                     uid, encoded_value)
            l = _LineItem(uid, receipt.authorisation())

            now2 = self._get_safe_now()

            if now2.hour == now.hour:
                # we are safely in the same hour
                break

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _credit(self, debit_note, bucket=None):
        """Credit the value in 'debit_note' to this account. If the debit_note
           shows that the payment is provisional then this will be recorded
           as accounts receivable. This will record the credit with the
           same UID as the debit identified in the debit_note, so that
           we can reconcile all credits against matching debits.

            Args:
                debit_note (DebitNote): Holds the value to be credited
                to this account
                TODO - improve bucket docs
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime): UID and current time
        """
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if debit_note.value() <= 0:
            return

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        if debit_note.is_provisional():
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.ACCOUNT_RECEIVABLE,
                                debit_note.value())
        else:
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.CREDIT,
                                debit_note.value())

        bucket = self._get_account_bucket()

        # create a UID and datetime for this credit and record
        # it in the account
        while True:
            now = self._get_safe_now()

            # and to create a key to find this credit later. The key is made
            # up from the isoformat datetime of the credit and a random string
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Accounting import LineItem as _LineItem
            from Acquire.ObjectStore import create_uuid as _create_uuid

            datetime_key = _datetime_to_string(now)
            uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

            item_key = "%s/%s/%s" % (self._transactions_key(),
                                     uid, encoded_value)

            now2 = self._get_safe_now()

            if now2.hour == now.hour:
                # we are safely in the same hour
                break

        # the line item records the UID of the debit note, so we can
        # find this debit note in the system and, from this, get the
        # original transaction in the transaction record
        l = _LineItem(debit_note.uid(), debit_note.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit(self, transaction, authorisation,
               is_provisional, receipt_by,
               authorisation_resource=None, bucket=None):
        """Debit the value of the passed transaction from this account based
           on the authorisation contained
           in 'authorisation'. This will create a unique ID (UID) for
           this debit and will return this together with the datetime of the
           debit. If this transaction 'is_provisional' then it will be
           recorded as a liability, which must be receipted before
           'receipt_by'. If 'receipt_by' is None, then this will
           automatically be 1 week in the future

           The UID will encode both the date of the debit and provide a random
           ID that together can be used to identify the transaction associated
           with this debit in the future.

           This will raise an exception if the debit cannot be completed, e.g.
           if the authorisation is invalid, if the debit exceeds a limit or
           there are insufficient funds in the account

           Note that this function is private as it should only be called
           by the DebitNote class

            Args:
                transaction (Transaction): Holds the value to be debited
                from this account
                authorisation (Authorisation): Authorisation for the
                transaction
                is_provisional (bool): If True the transaction will be
                recorded as a liability
                receipt_by (datetime): Datetime by which the transaction
                should be receipted
                TODO - improve bucket docs
                bucket (dict, default=None): Bucket to load data from

            Returns:
                tuple (str, datetime, datetime): uid, now, receipt_by

        """

        if self.is_null() or transaction.value() <= 0:
            return None

        from Acquire.Accounting import Transaction as _Transaction

        if not isinstance(transaction, _Transaction):
            raise TypeError("The passed transaction must be a Transaction!")

        if authorisation_resource is None:
            authorisation_resource = transaction.fingerprint()
            accept_partial_match = True
        else:
            accept_partial_match = False

        self.assert_valid_authorisation(
                                    authorisation=authorisation,
                                    resource=authorisation_resource,
                                    accept_partial_match=accept_partial_match)

        bucket = self._get_account_bucket()

        balance = self.balance(bucket=bucket)

        if balance.available(self.get_overdraft_limit()) < transaction.value():
            from Acquire.Accounting import InsufficientFundsError
            raise InsufficientFundsError(
                "You cannot debit '%s' from account %s as there "
                "are insufficient funds in this account." %
                (transaction, str(self)))

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        from Acquire.ObjectStore import get_datetime_future \
            as _get_datetime_future
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        while True:
            # create a UID and datetime for this debit and record
            # it in the account
            now = self._get_safe_now()

            if is_provisional:
                if receipt_by is None:
                    receipt_by = _get_datetime_future(days=7)
                else:
                    receipt_by = _datetime_to_datetime(receipt_by)

                delta = (receipt_by - now).total_seconds()
                if delta < 3600:
                    from Acquire.Accounting import AccountError
                    raise AccountError(
                        "You cannot request a receipt to be provided less "
                        "than 1 hour into the future! %s versus %s is only "
                        "%s second(s) in the future!" %
                        (_datetime_to_string(receipt_by),
                         _datetime_to_string(now), delta))
            else:
                receipt_by = None

            # and to create a key to find this debit later. The key is made
            # up from the isoformat datetime of the debit and a random string
            from Acquire.ObjectStore import create_uuid as _create_uuid
            datetime_key = _datetime_to_string(now)
            uid = "%s/%s" % (datetime_key, _create_uuid()[0:8])

            # the key in the object store is a combination of the key for this
            # account plus the uid for the debit plus the actual debit value.
            # We record the debit value in the key so that we can accumulate
            # the balance from just the key names
            if is_provisional:
                encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.CURRENT_LIABILITY,
                                    transaction.value())
            else:
                encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.DEBIT,
                                    transaction.value())

            item_key = "%s/%s/%s" % (self._transactions_key(),
                                     uid, encoded_value)

            # create a line_item for this debit and save it to the object store
            line_item = _LineItem(uid, authorisation)

            # validate that we have not stepped into another hour...
            now2 = self._get_safe_now()

            if now.hour == now2.hour:
                # we are still in the same hour, so it is safe to
                # record the transaction
                break

        _ObjectStore.set_object_from_json(bucket=bucket, key=item_key,
                                          data=line_item.to_data())

        balance = self.balance(bucket=bucket)

        if balance.available(overdraft_limit=self._overdraft_limit) < 0:
            # This transaction has helped push the account beyond the
            # overdraft limit. This can only happen if two debits
            # take place at the same time - both should be refunded
            from Acquire.Accounting import TransactionInfo \
                as _TransactionInfo

            info = _TransactionInfo.from_key(item_key)
            info = _TransactionInfo.rescind(info)

            line_item = _LineItem(uid=info.dated_uid(), authorisation=None)

            item_key = "%s/%s" % (self._transactions_key(),
                                  info.to_key())

            _ObjectStore.set_object_from_json(bucket=bucket, key=item_key,
                                              data=line_item.to_data())

            raise InsufficientFundsError(
                "You cannot debit '%s' from account %s as there "
                "are insufficient funds in this account." %
                (transaction, str(self)))

        return (uid, now, receipt_by)

    def get_overdraft_limit(self):
        """Return the overdraft limit of this account

            Returns:
                Decimal: Overdraft limit
        """
        if self.is_null():
            return 0

        return self._overdraft_limit

    def set_group(self, group, bucket=None):
        """Set the Accounts group to which this account belongs"""
        if self.is_null():
            return

        from Acquire.Accounting import Accounts as _Accounts
        if not isinstance(group, _Accounts):
            raise TypeError("The Accounts group must be of type Accounts")

        if self._group_name != group.name():
            self._group_name = group.name()
            self._save_account(bucket=bucket)

    def set_overdraft_limit(self, limit, bucket=None):
        """Set the overdraft limit of this account to 'limit'"""
        if self.is_null():
            return

        from Acquire.Accounting import create_decimal as _create_decimal
        limit = _create_decimal(limit)
        if limit < 0:
            raise ValueError("You cannot set the overdraft limit to a "
                             "negative value! (%s)" % limit)

        old_limit = self._overdraft_limit

        if old_limit != limit:
            self._overdraft_limit = limit

            if self.is_beyond_overdraft_limit():
                # restore the old limit
                self._overdraft_limit = old_limit
                from Acquire.Accounting import AccountError
                raise AccountError("You cannot change the overdraft limit to "
                                   "%s as this is greater than the current "
                                   "balance!" % (limit))
            else:
                # save the new limit to the object store
                self._save_account(bucket)

    def is_beyond_overdraft_limit(self, bucket=None):
        """Return whether or not the current balance is beyond
           the overdraft limit

            Args:
                TODO
                bucket (dict, default=None):
            Returns:
                bool: True if over overdraft limit, else False
        """
        available = self.balance(bucket=bucket).available()
        return available < -(self.get_overdraft_limit())

    def _key(self):
        """Return the key for this account in the object store"""
        if self.is_null():
            return None
        else:
            return "%s/%s" % (_account_root(), self.uid())

    def _transactions_key(self):
        """Return the root key for the transactions for this account
           in the object store
        """
        if self.is_null():
            return None
        else:
            return "%s/txns" % self._key()

    def _balance_key(self):
        """Return the root key for the balances for this account
           in this object store
        """
        if self.is_null():
            return None
        else:
            return "%s/balance" % self._key()

    def _load_account(self, bucket=None):
        """Load the current state of the account from the object store"""
        if self.is_null():
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = self._get_account_bucket()

        try:
            data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                     key=self._key())
        except:
            data = None

        if data is None:
            from Acquire.Accounting import AccountError
            raise AccountError(
                "There is no account data for this account? %s" % self._key())

        import copy as _copy
        self.__dict__ = _copy.copy(Account.from_data(data).__dict__)

    def _save_account(self, bucket=None):
        """Save this account back to the object store"""
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = self._get_account_bucket()
        _ObjectStore.set_object_from_json(bucket=bucket,
                                          key=self._key(),
                                          data=self.to_data())

        # reload, in case anyone saved just after us...
        self._load_account(bucket=bucket)

    def to_data(self):
        """Return a dictionary that can be encoded to json from this object"""
        data = {}

        if not self.is_null():
            data["uid"] = self._uid
            data["name"] = self._name
            data["description"] = self._description
            data["overdraft_limit"] = str(self._overdraft_limit)
            data["aclrules"] = self._aclrules.to_data()
            data["group_name"] = self._group_name

        return data

    @staticmethod
    def from_data(data):
        """Construct and return an Account from the passed dictionary that has
           been decoded from json
        """
        account = Account()

        if (data and len(data) > 0):
            from Acquire.Accounting import create_decimal as _create_decimal
            from Acquire.Identity import ACLRules as _ACLRules

            account._uid = data["uid"]
            account._name = data["name"]
            account._description = data["description"]
            account._overdraft_limit = _create_decimal(data["overdraft_limit"])

            if "aclrules" in data:
                account._aclrules = _ACLRules.from_data(data["aclrules"])
            else:
                account._aclrules = _ACLRules.inherit()

            if "group_name" in data:
                account._group_name = data["group_name"]
            else:
                account._group_name = None

        return account




__all__ = ["ChunkDownloader"]


class ChunkDownloader:
    """This class is used to control the chunked downloading
       of a file. This allows a file to be downloaded
       chunk by chunk (bit by bit). This is useful, e.g.
       to download a file as it is being written
    """
    def __init__(self, drive_uid=None, file_uid=None):
        """Create a new ChunkDowloader that downloads the specified
           file from the specified drive
        """
        self._uid = None
        self._drive_uid = None
        self._file_uid = None
        self._service = None

        self._next_index = None
        self._last_filename = None
        self._downloaded_filename = None
        self._FILE = None

        if drive_uid is not None:
            self._drive_uid = str(drive_uid)
            self._file_uid = str(file_uid)
            from Acquire.Crypto import PrivateKey as _PrivateKey
            self._secret = _PrivateKey.random_passphrase()

            from Acquire.ObjectStore import create_uid as _create_uid
            self._uid = _create_uid(short_uid=True)

    def __exit__(self, exception_type, exception_value, traceback):
        """Ensure that we close the file"""
        if self.is_open():
            self.close()

    def __del__(self):
        """Make sure that we close the file before deleting this object"""
        if self.is_open():
            self.close()

    def is_null(self):
        """Return whether or not this is null"""
        return self._uid is None

    def uid(self):
        """Return the UID of this downloader"""
        return self._uid

    def secret(self):
        """Return the secret used to authenticate the download"""
        return self._secret

    def service(self):
        """Return the service that created this downloader"""
        return self._service

    def local_filename(self):
        """Return the name of the local file to which we are downloading
           data
        """
        return self._downloaded_filename

    def _start_download(self, filename=None, dir=None):
        """Start the download of the file to 'filename' in 'dir'"""
        if self.is_null():
            raise PermissionError(
                "Cannot download a chunk using a null uploader!")

        if filename is None:
            filename = self._last_filename

        if filename is None:
            raise PermissionError(
                "You must supply a filename to which to download the file!")

        if self._last_filename is None:
            self._last_filename = filename
            self._next_index = 0
            from Acquire.Client import create_new_file as \
                _create_new_file
            self._downloaded_filename = _create_new_file(filename=filename,
                                                         dir=dir)
            self._FILE = open(self._downloaded_filename, "ab")
        elif self._last_filename != filename:
            raise PermissionError(
                "You cannot change the filename during an active "
                "streaming download!")

        return self._downloaded_filename

    def download_next_chunk(self):
        """Download the next chunk. Returns 'True' if something was
           downloaded, else it returns 'False'
        """
        if not self.is_open():
            return False

        service = self.service()

        if service is None:
            raise PermissionError(
                "Cannot download a chunk from a null service!")

        from Acquire.Crypto import Hash as _Hash

        secret = _Hash.multi_md5(self._secret,
                                 "%s%s%d" % (self._drive_uid,
                                             self._file_uid,
                                             self._next_index))

        args = {}
        args["uid"] = self._uid
        args["drive_uid"] = self._drive_uid
        args["file_uid"] = self._file_uid
        args["chunk_index"] = self._next_index
        args["secret"] = secret

        response = service.call_function(function="download_chunk",
                                         args=args)

        if "meta" in response:
            import json as _json
            meta = _json.loads(response["meta"])
            checksum = meta["checksum"]

            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes

            chunk = _string_to_bytes(response["chunk"])

            md5 = _Hash.md5(chunk)

            if checksum != md5:
                from Acquire.Storage import FileValidationError
                raise FileValidationError(
                    "Problem downloading - checksums don't agree: %s vs %s" %
                    (checksum, md5))

            import bz2 as _bz2
            chunk = _bz2.decompress(chunk)
            self._FILE.write(chunk)
            self._FILE.flush()
            chunk = None

            self._next_index = self._next_index + 1

        if "num_chunks" in response:
            num_chunks = int(response["num_chunks"])

            if self._next_index >= num_chunks:
                # nothing more to download
                self.close()

        return True

    def download(self, filename=None, dir=None):
        """Download as much of the file as possible to 'filename'. You
           can call this repeatedly with the same filename (or with
           no filename set) to stream the file back as it is written
        """
        self._start_download(filename=filename, dir=dir)
        downloaded_filename = self._downloaded_filename

        got_chunk = self.download_next_chunk()

        while got_chunk:
            next_index = self._next_index
            got_chunk = self.download_next_chunk()

            if got_chunk:
                assert(next_index != self._next_index)

        return downloaded_filename

    def is_open(self):
        """Return whether or not the file is open (has been written to)"""
        return self._next_index is not None

    def close(self):
        """Close the downloader"""
        if self.is_open():
            args = {"uid": self._uid,
                    "drive_uid": self._drive_uid,
                    "file_uid": self._file_uid,
                    "secret": self._secret}

            self.service().call_function(function="close_downloader",
                                         args=args)

            self._FILE.close()
            self._last_filename = None
            self._downloaded_filename = None
            self._next_index = None
            self._secret = None
            self._drive_uid = None
            self._file_uid = None
            self._uid = None

    def to_data(self, pubkey=None):
        """Return a json-serialisable dictionary of the object. If
           'pubkey' is supplied, then sensitive data used by the
           uploader is encrypted
        """
        if self.is_null():
            return {}

        data = {}

        data["drive_uid"] = self._drive_uid
        data["file_uid"] = self._file_uid
        data["secret"] = self._secret
        data["uid"] = self._uid

        if pubkey is not None:
            from Acquire.Crypto import PublicKey as _PublicKey
            from Acquire.ObjectStore import bytes_to_string \
                as _bytes_to_string
            import json as _json

            if not isinstance(pubkey, _PublicKey):
                raise TypeError("pubkey must be type PublicKey")

            d = _bytes_to_string(pubkey.encrypt(_json.dumps(data)))

            data = {}
            data["is_encrypted"] = True
            data["data"] = d

        return data

    @staticmethod
    def from_data(data, privkey=None, service=None):
        """Return a ChunkDownloader from a json-deserialised dictionary.
           If this was encrypted then you need to supply a private
           key to decrypt the sensitive data
        """
        if data is None or len(data) == 0:
            return ChunkDownloader()

        c = ChunkDownloader()

        try:
            is_encrypted = data["is_encrypted"]
        except:
            is_encrypted = False

        if is_encrypted:
            if privkey is None:
                raise PermissionError("Cannot decode as encrypted!")

            from Acquire.Crypto import PrivateKey as _PrivateKey
            if not isinstance(privkey, _PrivateKey):
                raise TypeError("privkey must be type PrivateKey")

            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes
            import json as _json
            data = _json.loads(privkey.decrypt(_string_to_bytes(data["data"])))

        c._uid = data["uid"]
        c._drive_uid = data["drive_uid"]
        c._file_uid = data["file_uid"]
        c._secret = data["secret"]
        c._service = service

        return c




__all__ = ["Registry"]

_registry_key = "registry"


def _inc_uid(vals):
    import copy as _copy
    vals = _copy.copy(vals)

    for j in range(0, len(vals)):
        i = len(vals) - j - 1
        vals[i] += 1

        if i % 2 == 1:
            if vals[i] < 10:
                return vals
            else:
                vals[i] = 0
        else:
            if vals[i] < 52:
                return vals
            else:
                vals[i] = 0

    # we only get here if we have overflowed. In this
    # case, add an extra pair of digits
    vals = [0] * (len(vals) + 2)

    return vals


def _to_uid(vals):
    import string as _string
    parts = []
    for i in range(0, len(vals)):
        x = vals[i]
        if i % 2 == 1:
            if x < 0 or x > 9:
                raise ValueError(x)
            else:
                parts.append(str(x))
        else:
            if x < 0 or x > 51:
                raise ValueError(x)
            elif x < 26:
                parts.append(_string.ascii_lowercase[x])
            else:
                parts.append(_string.ascii_uppercase[x-26])

    s = "".join(parts)
    return ".".join([s[i:i+2] for i in range(0, len(s), 2)])


def _generate_service_uid(bucket, registry_uid):
    """Function to generate a new service_uid on this registry.

       The UIDs have the form a0-a0, when "a" is any letter from [a-zA-Z]
       and "0" is any number from [0-9]. This give 520 possible values
       for each part either side of the hyphen.

       The part on the left of the hypen is the root UID, which
       matches the root of the service_uid of the registry service
       that registered this service (the service_uid of a registry
       service has the UID root-root).

       If more than 520 values are needed, then either side of the
       ID can be extended by additional pairs of a0 digits, using
       a "." to separate pairs, e.g.

       the service_uid for registry b4-b4 that comes after
       b4-Z9.Z9.Z9 is b4-a0.a0.a0.a0

       similarly, the registry after Z9 is A0-A0.

       This means that

       a0.a0-a0.a0.a0.a0

       would be a perfectly valid ID. We would only need IDs of this
       length if we have ~270k registry services, and this service_uid
       came from a service that had registered ~73 trillion services...

       The registry root Z9, with registry Z9-Z9 is reserved for
       the temporary registry created during testing
    """
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.ObjectStore import Mutex as _Mutex

    root = registry_uid.split("-")[0]

    key = "%s/last_service_uid" % _registry_key

    mutex = _Mutex(key=key)

    try:
        last_vals = _ObjectStore.get_object_from_json(bucket=bucket,
                                                      key=key)
        last_vals = _inc_uid(last_vals)
    except:
        last_vals = [0, 0]

    service_uid = "%s-%s" % (root, _to_uid(last_vals))

    while service_uid == registry_uid:
        last_vals = _inc_uid(last_vals)
        service_uid = "%s-%s" % (root, _to_uid(last_vals))

    _ObjectStore.set_object_from_json(bucket=bucket, key=key, data=last_vals)
    mutex.unlock()

    return service_uid


class Registry:
    """This class holds the registry of all services registered by
       this service. Registries provided trusted actors who
       can supply public keys, URLs, and UIDs for all of the different
       services in the system.
    """
    def __init__(self):
        """Constructor"""
        self._bucket = None

        from Acquire.Service import get_this_service as _get_this_service
        self._registry_uid = _get_this_service(
                                need_private_access=False).uid()

    def get_registry_uid(self, service_uid):
        """Return the UID of the registry that initially registered
           the service with UID 'service_uid'.
        """
        parts = service_uid.split("-")
        return "%s-%s" % (parts[0], parts[0])

    def get_bucket(self):
        if self._bucket:
            return self._bucket
        else:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            self._bucket = _get_service_account_bucket()
            return self._bucket

    def registry_root(self):
        """Return the root ID of this registry"""
        return self._registry_uid.split("-")[0]

    def registry_uid(self):
        """Return the UID of this registry"""
        return self._registry_uid

    def is_test_registry(self):
        """Return whether or not this is a test registry (only used
           while testing the code)
        """
        return self._registry_uid == "Z9-Z9"

    def _get_key_for_uid(self, service_uid):
        return "%s/uid/%s" % (_registry_key, service_uid)

    def _get_uid_from_key(self, key):
        return key.split("/")[-1]

    def _get_domain(self, url):
        """Return the domain name of the server in the passed url"""
        from urllib.parse import urlparse as _urlparse
        try:
            domain = _urlparse(url).netloc.split(":")[0]
        except:
            domain = None

        if domain is None or len(domain) == 0:
            if self.is_test_registry():
                return "testing"
            else:
                raise ValueError("Cannot extract the domain from '%s'"
                                 % url)

        return domain

    def _get_root_key_for_domain(self, domain):
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        return "%s/domain/%s" % (_registry_key, _string_to_encoded(domain))

    def _get_key_for_url(self, service_url):
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "%s/url/%s" % (_registry_key, _string_to_encoded(service_url))

    def get_service_key(self, service_uid=None, service_url=None):
        """Return the key for the passed service in the object store"""
        if service_uid is not None:
            return self._get_key_for_uid(service_uid)
        else:
            bucket = self.get_bucket()
            key = self._get_key_for_url(service_url)

            try:
                from Acquire.ObjectStore import ObjectStore as _ObjectStore
                service_key = _ObjectStore.get_string_object(bucket=bucket,
                                                             key=key)
            except:
                service_key = None

            return service_key

    def challenge_service(self, service):
        """Send a challenge to the passed service, returning the actual
           service returned. This will only pass if our copy of the
           service matches us with the copy returned from the actual
           service. This verifies that there is a real service sitting
           at that URL, and that we have the right keys and certs
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        from Acquire.Service import Service as _Service

        challenge = _PrivateKey.random_passphrase()
        pubkey = service.public_key()
        encrypted_challenge = pubkey.encrypt(challenge)

        args = {"challenge": _bytes_to_string(encrypted_challenge),
                "fingerprint": pubkey.fingerprint()}

        if service.uid().startswith("STAGE"):
            # we cannot call using the service directly, as it is
            # not yet fully constructed
            from Acquire.Service import get_this_service as _get_this_service
            from Acquire.Service import call_function as _call_function
            this_service = _get_this_service(need_private_access=True)
            result = _call_function(service_url=service.service_url(),
                                    function=None,
                                    args=args,
                                    args_key=service.public_key(),
                                    response_key=this_service.private_key(),
                                    public_cert=service.public_certificate())
        else:
            result = service.call_function(function=None, args=args)

        if result["response"] != challenge:
            raise PermissionError(
                "Failure of the service %s to correctly respond "
                "to the challenge!" % service)

        return _Service.from_data(result["service_info"])

    def get_service(self, service_uid=None, service_url=None):
        """Load and return the service with specified url or uid
           from the registry. This will consult with other
           registry services to find the matching service
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import Service as _Service
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        from Acquire.Service import get_this_service as _get_this_service

        this_service = _get_this_service(need_private_access=False)

        if service_url is not None:
            from Acquire.Service import Service as _Service
            service_url = _Service.get_canonical_url(service_url)

        if this_service.uid() == service_uid:
            return this_service
        elif this_service.canonical_url() == service_url:
            return this_service

        bucket = self.get_bucket()

        service_key = self.get_service_key(service_uid=service_uid,
                                           service_url=service_url)

        service = None

        if service_key is not None:
            try:
                data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                         key=service_key)
                service = _Service.from_data(data)
            except:
                pass

        if service is not None:
            must_write = False

            if service.uid() == "STAGE1":
                # we need to directly ask the service for its info
                service = self.challenge_service(service)

                if service.uid() == "STAGE1":
                    from Acquire.Service import MissingServiceError
                    raise MissingServiceError(
                        "Service %s|%s not available as it is still under "
                        "construction!" % (service_uid, service))

                # we can now move this service from pending to active
                uidkey = self._get_key_for_uid(service.uid())
                domain = self._get_domain(service.service_url())
                domainroot = self._get_root_key_for_domain(domain=domain)

                pending_key = "%s/pending/%s" % (domainroot, service.uid())
                active_key = "%s/active/%s" % (domainroot, service.uid())

                try:
                    _ObjectStore.delete_object(bucket=bucket,
                                               key=pending_key)
                except:
                    pass

                try:
                    _ObjectStore.set_string_object(bucket=bucket,
                                                   key=active_key,
                                                   string_data=uidkey)
                except:
                    pass

                must_write = True
            elif service.should_refresh_keys():
                service.refresh_keys()
                must_write = True

            if must_write:
                data = service.to_data()
                _ObjectStore.set_object_from_json(bucket=bucket,
                                                  key=service_key,
                                                  data=data)
            return service

        # we only get here if we can't find the service on this registry.
        # In the future, we will use the initial part of the UID of
        # the service to ask its registering registry for its data.
        # For now, we just raise an error
        from Acquire.Service import MissingServiceError
        raise MissingServiceError(
            "No service available: service_url=%s  service_uid=%s" %
                                  (service_url, service_uid))

    def register_service(self, service, force_new_uid=False):
        """Register the passed service"""
        from Acquire.Service import Service as _Service
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        if not isinstance(service, _Service):
            raise TypeError("You can only register Service objects")

        if service.uid() != "STAGE1":
            raise PermissionError("You cannot register a service twice!")

        # first, stop a single domain monopolising resources...
        bucket = self.get_bucket()
        domain = self._get_domain(service.service_url())
        domainroot = self._get_root_key_for_domain(domain=domain)

        try:
            pending_keys = _ObjectStore.get_all_object_names(
                                        bucket=bucket,
                                        prefix="%s/pending/" % domainroot)
            num_pending = len(pending_keys)
        except:
            num_pending = 0

        if num_pending >= 4:
            raise PermissionError(
                "You cannot register a new service as you have reached "
                "the quota (4) for the number of pending services registered "
                "against the domain '%s'. Please get some of these services "
                "so that you can make them active." % domain)

        try:
            active_keys = _ObjectStore.get_all_object_names(
                                        bucket=bucket,
                                        prefix="%s/active/" % domainroot)
            num_active = len(active_keys)
        except:
            num_active = 0

        if num_active + num_pending >= 16:
            raise PermissionError(
                "You cannot register a new service as you have reached "
                "the quota (16) for the number registered against the "
                "domain '%s'" % domain)

        # first, challenge the service to ensure that it exists
        # and our keys are correct
        service = self.challenge_service(service)

        if service.uid() != "STAGE1":
            raise PermissionError("You cannot register a service twice!")

        bucket = self.get_bucket()
        urlkey = self._get_key_for_url(service.canonical_url())

        try:
            uidkey = _ObjectStore.get_string_object(bucket=bucket,
                                                    key=urlkey)
        except:
            uidkey = None

        service_uid = None

        if uidkey is not None:
            # there is already a service registered at this domain. Since
            # we have successfully challenged the service, this must be
            # someone re-bootstrapping a service. It is safe to give them
            # back their UID if requested
            if not force_new_uid:
                service_uid = self._get_uid_from_key(uidkey)

        if service_uid is None:
            # how many services from this domain are still pending?

            service_uid = _generate_service_uid(
                                        bucket=self.get_bucket(),
                                        registry_uid=self.registry_uid())

        # save this service to the object store
        uidkey = self._get_key_for_uid(service_uid)

        _ObjectStore.set_object_from_json(bucket=bucket, key=uidkey,
                                          data=service.to_data())

        _ObjectStore.set_string_object(bucket=bucket, key=urlkey,
                                       string_data=uidkey)

        domainkey = self._get_root_key_for_domain(domain=domain)

        _ObjectStore.set_string_object(
                                bucket=bucket,
                                key="%s/pending/%s" % (domainkey, service_uid),
                                string_data=uidkey)

        return service_uid

    def _get_key_for_uid(self, service_uid):
        return "%s/uid/%s" % (_registry_key, service_uid)

    def _get_key_for_url(self, service_url):
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "%s/url/%s" % (_registry_key, _string_to_encoded(service_url))




__all__ = ["register_service"]


def register_service(service, registry_uid):
    """Call this function to register the passed new service with
       the specified registry_uid. This function will complete
       registration and construction of the service
    """
    from Acquire.Service import Service as _Service
    if not isinstance(service, _Service):
        raise TypeError("You can only register Service objects")

    if not service.uid().startswith("STAGE1"):
        raise PermissionError(
            "You can only register services that are at STAGE1 of "
            "construction")

    if service.service_type() == "registry":
        from Acquire.Registry import get_registry_details \
            as _get_registry_details

        details = _get_registry_details(registry_uid=registry_uid)

        from Acquire.Service import Service as _Service
        canonical_url = _Service.get_canonical_url(details["canonical_url"])

        # make sure that everything matches what was specified
        if canonical_url != service.canonical_url():
            raise PermissionError(
                "Cannot change the canonical URL. I expect %s, but "
                "you are trying to set to %s" %
                (service.canonical_url(), details["canonical_url"]))

        from Acquire.Registry import update_registry_keys_and_certs \
            as _update_registry_keys_and_certs

        _update_registry_keys_and_certs(
                            registry_uid=registry_uid,
                            public_key=service.public_key(),
                            public_certificate=service.public_certificate())

        service.create_stage2(service_uid=registry_uid,
                              response=service._uid)
        return service

    # get the trusted registry
    from Acquire.Registry import get_trusted_registry_service \
        as _get_trusted_registry_service
    registry_service = _get_trusted_registry_service(
                                        service_uid=registry_uid)

    if not registry_service.is_registry_service():
        raise PermissionError(
            "You can only register new services on an existing and valid "
            "registry service. Not %s" % registry_service)

    from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
    pubkey = registry_service.public_key()
    challenge = pubkey.encrypt(service.uid())

    args = {"service": service.to_data(),
            "challenge": _bytes_to_string(challenge),
            "fingerprint": pubkey.fingerprint()}

    result = registry_service.call_function(function="register_service",
                                            args=args)

    service_uid = result["service_uid"]
    response = result["response"]

    service.create_stage2(service_uid=service_uid, response=response)

    return service





__all__ = ["WeakPassphraseError", "KeyManipulationError",
           "SignatureVerificationError",
           "DecryptionError", "OTPError",
           "RepeatedOTPCodeError"]


class WeakPassphraseError(Exception):
    pass


class KeyManipulationError(Exception):
    pass


class SignatureVerificationError(Exception):
    pass


class DecryptionError(Exception):
    pass


class OTPError(Exception):
    pass


class RepeatedOTPCodeError(OTPError):
    pass




import uuid as _uuid
from copy import copy as _copy

from Acquire.Service import Service as _Service

__all__ = ["AccessService"]


class AccessService(_Service):
    """This is a specialisation of Service for Access Services
       Where Service represents a service in the system. Services
       will either be identity services, access services,
       storage services or accounting services
    """

    def __init__(self, other=None):
        if isinstance(other, _Service):
            self.__dict__ = _copy(other.__dict__)

            if not self.is_access_service():
                from Acquire.Access import AccessServiceError
                raise AccessServiceError(
                    "Cannot construct an AccessService from "
                    "a service which is not an access service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls

           Args:
                function: function to route
                args: arguments to be passed into function

            Returns:
                function: function object
        """

        from access.route import access_functions as _access_functions
        from admin.handler import create_handler as _create_handler

        handler = _create_handler(_access_functions)
        return handler(function=function, args=args)

    def get_trusted_storage_service(self):
        """Return a trusted storage service

            Returns:
                :obj:`dict`: containing the first storage
                device on the trusted service

        """
        from Acquire.Service import get_trusted_services \
            as _get_trusted_services

        services = _get_trusted_services()

        try:
            return services["storage"][0]
        except:
            from Acquire.Service import ServiceError
            raise ServiceError(
                "There is no trusted storage service known to this access "
                "service")




__all__ = ["LoginSession"]

_sessions_key = "identity/sessions"


class LoginSession:
    """This class holds all details of a single login session"""
    def __init__(self, username=None,
                 public_key=None, public_cert=None, ipaddr=None,
                 hostname=None, login_message=None, scope=None,
                 permissions=None):
        """Start a new login session for the user with specified
           username, passing in the additional data needed to
           request a login
        """
        if public_key is not None:
            from Acquire.Crypto import PublicKey as _PublicKey
            if not isinstance(public_key, _PublicKey):
                raise TypeError("The public key must be of type PublicKey")

            if not isinstance(public_cert, _PublicKey):
                raise TypeError("The public certificate must be of "
                                "type PublicKey")

            if username is None or len(username) == 0:
                raise PermissionError("You must supply a valid username!")

            self._username = username
            self._pubkey = public_key
            self._pubcert = public_cert

            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            from Acquire.ObjectStore import create_uuid as _create_uuid

            self._uid = _create_uuid()
            self._request_datetime = _get_datetime_now()
            self._status = None

            self._ipaddr = ipaddr
            self._hostname = hostname
            self._login_message = login_message
            self._scope = scope
            self._permissions = permissions

            # make sure this session is saved to the object store
            self._set_status("pending")
        else:
            self._uid = None

    def __str__(self):
        if self.is_null():
            return "LoginSession::null"
        else:
            return "LoginSession(uid=%s, status=%s)" % \
                            (self.uid(), self.status())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this object is null"""
        return self._uid is None

    def username(self):
        """Return the username for this login session"""
        try:
            return self._username
        except:
            return None

    def encoded_username(self):
        """Return a safely-encoded version of the username. This is used
           to create safe keys in the object store
        """
        if self.is_null():
            raise PermissionError(
                "You cannot get the encoded username of a null LoginSession")

        from Acquire.Identity import _encode_username
        return _encode_username(self._username)

    def public_key(self):
        """Return the public key"""
        status = self.status()

        if status != "approved":
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot get a public key from "
                "a session that is not fully approved (status = %s)" % status)

        return self._pubkey

    def public_certificate(self):
        """Return the public certificate"""
        status = self.status()

        if status != "approved":
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot get a public certificate from "
                "a session that is not fully approved (status = %s)" % status)

        return self._pubcert

    def request_source(self):
        """Return the IP address of the source of
           this request. This could be used to rate limit someone
           who is maliciously requesting logins...
        """
        if self.is_null():
            return None

        return self._ipaddr

    def uid(self):
        """Return the UID of this request"""
        if self.is_null():
            return None

        return self._uid

    @staticmethod
    def to_short_uid(long_uid):
        """Return the short UID version of the passed long uid"""
        return long_uid[:8]

    def short_uid(self):
        """Return a short UUID that will be used to
           provide a more human-readable session ID
        """
        if self._uid:
            return LoginSession.to_short_uid(self._uid)
        else:
            return None

    def login_url(self):
        """Return the login URL to login to this session. This is
           the URL of this identity service plus the
           short UID of the session
        """
        from Acquire.Service import get_this_service as _get_this_service
        service = _get_this_service(need_private_access=False)
        service_uid = service.uid()
        short_uid = self.short_uid()
        short_uid = ".".join([short_uid[i:i+2]
                             for i in range(0, len(short_uid), 2)])
        # eventually allow the service provider to configure this url
        url = "https://login.acquire-aaai.com"

        return "%s?id=%s/%s" % (url, service_uid, short_uid)

    def regenerate_uid(self):
        """Regenerate the UUID as there has been a clash"""
        if not self.is_null():
            from Acquire.ObjectStore import create_uuid as _create_uuid
            self._uid = _create_uuid()

    def creation_time(self):
        """Return the date and time when this was created"""
        if self.is_null():
            return None

        return self._request_datetime

    def login_time(self):
        """Return the date and time when the user logged in. This
           returns None if the user has not yet logged in
        """
        try:
            return self._login_datetime
        except:
            return None

    def logout_time(self):
        """Return the date and time when the user logged out. This
           returns None if the user has not yet logged out
        """
        try:
            return self._logout_datetime
        except:
            return None

    def seconds_since_creation(self):
        """Return the number of seconds since this request was
           created
        """
        if self.is_null():
            return None

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now
        delta = _get_datetime_now() - self._request_datetime
        return delta.total_seconds()

    def status(self):
        """Return the status of this login session"""
        if self.is_null():
            return "null"
        else:
            return self._status

    @staticmethod
    def get_status(uid):
        """Return the status of the LoginSession with specified UID"""
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()

        try:
            key = "%s/status/%s" % (_sessions_key, uid)
            status = _ObjectStore.get_string_object(bucket=bucket, key=key)
        except:
            status = None

        if status is None:
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "Cannot find a session with UID '%s'" % uid)

        return status

    def _set_status(self, status):
        """Internal function to set the status of the session.
           This ensures that the data for the session is saved
           into the correct part of the object store
        """
        if self.is_null():
            raise PermissionError(
                "Cannot set the status of a null LoginSession")

        if status not in ["approved", "pending", "denied",
                          "suspicious", "logged_out"]:
            raise ValueError("Cannot set an invalid status '%s'" % status)

        if status == self._status:
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = self._get_key()

        try:
            _ObjectStore.delete_object(bucket=bucket, key=key)
        except:
            pass

        self._status = status
        key = self._get_key()
        _ObjectStore.set_object_from_json(bucket=bucket, key=key,
                                          data=self.to_data())
        key = "%s/status/%s" % (_sessions_key, self._uid)
        _ObjectStore.set_string_object(bucket=bucket, key=key,
                                       string_data=status)

    def set_suspicious(self):
        """Put this login session into a suspicious state. This
           will be because weird activity has been detected which indicates
           that the session may be have been cracked. A login session
           in a suspicious state should not be granted any permissions.
        """
        if not self.is_null():
            self._set_status("suspicious")

    def set_approved(self, user_uid=None, device_uid=None):
        """Register that this request has been approved, optionally
           providing data about the user who approved the session
           and the device from which the session was approved
        """
        if self.is_null():
            raise PermissionError(
                "You cannot approve a null LoginSession!")

        if self.status() != "pending":
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "You cannot approve a login session "
                "that is not in the 'unapproved' state. This login "
                "session is in the '%s' state." % self.status())

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now
        self._login_datetime = _get_datetime_now()
        self._user_uid = user_uid
        self._device_uid = device_uid

        self._set_status("approved")

    def _clear_keys(self):
        """Function called to remove all keys
           from this session, as it has now been terminated
           (and so the keys are no longer valid)
        """
        self._pubkey = None

    def set_denied(self):
        """Register that this request has been denied"""
        if self.is_null():
            raise PermissionError(
                "You cannot deny a null LoginSession!")

        self._clear_keys()
        self._pubcert = None
        self._set_status("denied")

    def set_logged_out(self, authorisation=None, signature=None):
        """Register that this request has been closed as
           the user has logged out. If an authorisation is
           passed then verify that this is correct
        """
        if self.is_null():
            raise PermissionError(
                "You cannot logout from a null LoginSession!")

        if authorisation is not None:
            from Acquire.Identity import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError("The authorisation must be type Authorisation")

            authorisation.verify(resource="logout %s" % self.uid())

            if authorisation.user_uid() != self.user_uid():
                raise PermissionError(
                    "The user '%s' does not have permission to logout "
                    "a session owned by %s" % (authorisation.user_uid(),
                                               self.user_uid()))

        if signature is not None:
            message = "logout %s" % self.uid()
            self._pubcert.verify(signature=signature, message=message)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        self._logout_datetime = _get_datetime_now()
        self._clear_keys()
        self._set_status("logged_out")

    def login(self, user_uid=None, device_uid=None):
        """Convenience function to set the session into the logged in state"""
        self.set_approved(user_uid=user_uid, device_uid=device_uid)

    def logout(self, authorisation=None, signature=None):
        """Convenience function to set the session into the logged out state"""
        self.set_logged_out(authorisation=authorisation,
                            signature=signature)

    def is_suspicious(self):
        """Return whether or not this session is suspicious"""
        return self.status() == "suspicious"

    def is_approved(self):
        """Return whether or not this session is open and
           approved by the user"""
        return self.status() == "approved"

    def is_logged_out(self):
        """Return whether or not this session has logged out"""
        return self.status() == "logged_out"

    def login_message(self):
        """Return any login message that has been set by the user"""
        try:
            return self._login_message
        except:
            return None

    def scope(self):
        """Return the scope requested for this login session"""
        try:
            return self._localised_scope
        except:
            try:
                return self._scope
            except:
                return None

    def permissions(self):
        """Return the permissions requested for this login session"""
        try:
            return self._localised_permissions
        except:
            try:
                return self._permissions
            except:
                return None

    def hostname(self):
        """Return the user-supplied hostname of the host making the
           login request
        """
        try:
            return self._hostname
        except:
            return None

    def ipaddr(self):
        """Return the user-supplied IP address of the host making the
           login request
        """
        try:
            return self._ipaddr
        except:
            return None

    def user_uid(self):
        """If known, return the UID of the user who approved this
           session
        """
        try:
            return self._user_uid
        except:
            return None

    def device_uid(self):
        """If known, return the UID of the device from which the
           user approved this session
        """
        try:
            return self._device_uid
        except:
            return None

    def _get_key(self):
        return "%s/%s/%s/%s" % (_sessions_key, self.status(),
                                self.short_uid(), self.uid())

    def save(self):
        """Save the current state of this LoginSession to the
           object store
        """
        if self.is_null():
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = self._get_key()

        _ObjectStore.set_object_from_json(bucket=bucket, key=key,
                                          data=self.to_data())

    def _localise(self, scope, permissions):
        """Localise this session for the specified scope and permissions.
           This will raise an exception if the scope or permissions are
           greater than those embedded into the session
        """
        # DO SOME WORK HERE TO ASSERT THAT THE SCOPE AND PERMISSIONS
        # ARE ACCEPTABLE

        if scope is None:
            self._localised_scope = self.scope()
        else:
            self._localised_scope = scope

        if permissions is None:
            self._localised_permissions = self.permissions()
        else:
            self._localised_permissions = permissions

    @staticmethod
    def load(status=None, short_uid=None, uid=None,
             scope=None, permissions=None):
        """Load and return a LoginSession specified from either a
           short_uid or a long uid. Note that if more than one
           session matches the short_uid then you will get a list
           of LoginSessions returned
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        if short_uid is None and uid is None:
            raise PermissionError(
                "You must supply one of the short_uid or uid to load "
                "a LoginSession!")

        if status is None:
            if uid is None:
                raise PermissionError(
                    "You must supply the full UID to get the status "
                    "of a specific login session")

            status = LoginSession.get_status(uid=uid)
        else:
            if status not in ["approved", "pending", "denied",
                              "suspicious", "logged_out"]:
                raise ValueError("Cannot set an invalid status '%s'" % status)

        bucket = _get_service_account_bucket()

        if uid is not None:
            short_uid = LoginSession.to_short_uid(uid)

            key = "%s/%s/%s/%s" % (_sessions_key, status, short_uid, uid)
            try:
                data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                         key=key)
                session = LoginSession.from_data(data)
            except:
                from Acquire.Identity import LoginSessionError
                raise LoginSessionError(
                    "There is no valid session with UID %s in "
                    "state %s" % (uid, status))

            session._localise(scope=scope, permissions=permissions)
            return session

        # the user may pass in the short_uid as YY.YY.YY.YY
        # so remove all dots
        short_uid = short_uid.replace(".", "")

        prefix = "%s/%s/%s/" % (_sessions_key, status, short_uid)

        try:
            keys = _ObjectStore.get_all_objects_from_json(bucket=bucket,
                                                          prefix=prefix)
        except:
            keys = {}

        sessions = []

        for data in keys.values():
            try:
                session = LoginSession.from_data(data)
                session._localise(scope=scope, permissions=permissions)
                sessions.append(session)
            except:
                pass

        if len(sessions) == 0:
            from Acquire.Identity import LoginSessionError
            raise LoginSessionError(
                "There is no valid session with short UID %s "
                "in state %s" % (short_uid, status))
        elif len(sessions) == 1:
            sessions = sessions[0]

        return sessions

    def to_data(self):
        """Return a data version (dictionary) of this LoginSession
           that can be serialised to json
        """
        if self.is_null():
            return {}

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        data = {}
        data["uid"] = self._uid
        data["username"] = self._username
        data["request_datetime"] = _datetime_to_string(self._request_datetime)
        data["public_certificate"] = self._pubcert.to_data()
        data["status"] = self._status

        if self._pubkey is not None:
            data["public_key"] = self._pubkey.to_data()

        try:
            data["login_datetime"] = _datetime_to_string(self._login_datetime)
        except:
            pass

        try:
            data["logout_datetime"] = _datetime_to_string(
                                            self._logout_datetime)
        except:
            pass

        try:
            data["ipaddr"] = self._ipaddr
        except:
            pass

        try:
            data["hostname"] = self._hostname
        except:
            pass

        try:
            data["login_message"] = self._login_message
        except:
            pass

        try:
            data["scope"] = self._scope
        except:
            pass

        try:
            data["permissions"] = self._permissions
        except:
            pass

        try:
            data["user_uid"] = self._user_uid
        except:
            pass

        try:
            data["device_uid"] = self._device_uid
        except:
            pass

        return data

    @staticmethod
    def from_data(data):
        """Return a LoginSession constructed from the passed data
           (dictionary)
        """
        l = LoginSession()

        if data is not None and len(data) > 0:
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            from Acquire.Crypto import PublicKey as _PublicKey

            l._uid = data["uid"]
            l._username = data["username"]
            l._request_datetime = _string_to_datetime(
                                        data["request_datetime"])
            l._pubcert = _PublicKey.from_data(data["public_certificate"])
            l._status = data["status"]

            try:
                l._pubkey = _PublicKey.from_data(data["public_key"])
            except:
                l._pubkey = None

            try:
                l._login_datatime = _string_to_datetime(
                                            data["login_datetime"])
            except:
                pass

            try:
                l._logout_datetime = _string_to_datetime(
                                            data["logout_datetime"])
            except:
                pass

            try:
                l._ipaddr = data["ipaddr"]
            except:
                pass

            try:
                l._hostname = data["hostname"]
            except:
                pass

            try:
                l._login_message = data["login_message"]
            except:
                pass

            try:
                l._scope = data["scope"]
            except:
                pass

            try:
                l._permissions = data["permissions"]
            except:
                pass

            try:
                l._user_uid = data["user_uid"]
            except:
                pass

            try:
                l._device_uid = data["device_uid"]
            except:
                pass

        return l




from enum import Enum as _Enum

__all__ = ["Cluster"]


class JobState(_Enum):
    PENDING = "pending"
    SUBMITTING = "submitting"
    STARTING = "starting"
    RUNNING = "running"
    ERROR = "error"
    TERMINATED = "terminated"
    COMPLETED = "completed"


class Cluster:
    """This class provides a handle to the unique compute
       cluster associated with a Compute service. There is only
       one cluster associated with a service, and this
       class provides the handle and functions used to communicate
       between the compute function service and the actual
       compute cluster.

       This class has two sides. On the compute service it provides
       the means to communicate with the cluster. On the cluster
       it provides the means to manage jobs and communicate with
       the compute service.
    """
    def __init__(self):
        self._uid = None

    def is_null(self):
        return self._uid is None

    def fingerprint(self):
        """Return a fingerprint of this cluster, suitable for
           authorisations
        """
        return self.uid()

    def passphrase(self, resource):
        """Generate a passphrase for the passed resource. This is used
           by the client running on the cluster to validate that it is
           allowed to communicate with the cluster on the service
        """
        if self.is_null():
            raise PermissionError(
                "You cannot generate a passphrase from a null Cluster")

        from Acquire.Crypto import Hash as _Hash
        return _Hash.multi_md5(self._secret, resource)

    def verify_passphrase(self, resource, passphrase):
        """Verify that the passed passphrase is correct for the specified
           resource. This is used on the service to validate that a request
           for a resource came from the Cluster running on the client
        """
        if self.is_null():
            raise PermissionError(
                "Invalid passphrase as this is a null Cluster")

        if passphrase != self.passphrase(resource):
            raise PermissionError(
                "Invalid passphrase for resource %s" % resource)

    def compute_service(self):
        """Return the compute service from which this cluster is managed"""
        if self.is_null():
            return None
        elif Cluster._is_running_service():
            from Acquire.Service import get_this_service as _get_this_service
            return _get_this_service()
        else:
            return self._compute_service

    @staticmethod
    def _is_running_service():
        """Return whether or not this is the Cluster on the compute
           service (otherwise it must be on the client)
        """
        from Acquire.Service import is_running_service as _is_running_service
        return _is_running_service()

    @staticmethod
    def create(service_url=None, service_uid=None, user=None):
        """Create a new cluster"""
        if Cluster._is_running_service():
            raise PermissionError(
                "You cannot create a Cluster on a running service")

        from Acquire.Client import PrivateKey as _PrivateKey
        from Acquire.ObjectStore import create_uid as _create_uid
        from Acquire.Client import Wallet as _Wallet

        wallet = _Wallet()
        compute_service = wallet.get_service(service_url=service_url,
                                             service_uid=service_uid)

        if not compute_service.is_compute_service():
            raise TypeError(
                "You can only create a cluster that will communicate "
                "with a valid compute service - not %s" % compute_service)

        cluster = Cluster()
        cluster._uid = _create_uid()
        cluster._private_key = _PrivateKey()
        cluster._public_key = cluster._private_key.public_key()
        cluster._compute_service = compute_service
        cluster._secret = _PrivateKey.random_passphrase()
        cluster._oldkeys = []

        if user is not None:
            Cluster.set_cluster(cluster=cluster, user=user)

        return cluster

    @staticmethod
    def get_cluster():
        """Return a handle to the single compute cluster that is
           connected to this compute service
        """
        if not Cluster._is_running_service():
            raise PermissionError(
                "You can only call 'get_cluster' on the compute service")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()
        key = "compute/cluster"

        try:
            data = _ObjectStore.get_object_from_json(bucket, key)
        except:
            data = None

        if data is None:
            from Acquire.Service import ServiceError
            raise ServiceError(
                "You have not set the cluster that will be used to actually "
                "run the compute jobs!")

        return Cluster.from_data(data)

    @staticmethod
    def set_cluster(cluster, authorisation=None, passphrase=None, user=None):
        """Function used to set the single compute cluster that is
           connected to this compute service. This must be authorised
           by an admin user of this compute service
        """
        if not isinstance(cluster, Cluster):
            raise TypeError("The cluster must be type Cluster")

        resource = "set_cluster %s" % cluster.fingerprint()

        from Acquire.Client import Authorisation as _Authorisation
        if Cluster._is_running_service():
            from Acquire.Service import get_this_service as _get_this_service

            service = _get_this_service(need_private_access=True)

            if authorisation is not None:
                if not isinstance(authorisation, _Authorisation):
                    raise TypeError(
                        "The authorisation must be type Authorisation")

                service.assert_admin_authorised(authorisation, resource)
            else:
                # we are rotating keys, so check the passphrase against
                # the old passphrase
                cluster = Cluster.get_cluster()
                cluster.verify_passphrase(passphrase=passphrase,
                                          resource="set_cluster")

            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket

            bucket = _get_service_account_bucket()
            key = "compute/cluster"
            _ObjectStore.set_object_from_json(bucket, key, cluster.to_data())
        else:
            authorisation = _Authorisation(user=user, resource=resource)
            compute_service = cluster.compute_service()

            args = {"authorisation": authorisation.to_data(),
                    "cluster": cluster.to_data()}

            compute_service.call_function(function="set_cluster", args=args)

    def uid(self):
        """Return the UID of this cluster"""
        return self._uid

    def public_key(self):
        """Return the public encryption key for this cluster"""
        if self.is_null():
            return None
        else:
            return self._public_key

    def private_key(self):
        """Return the private key for this cluster (if available)"""
        if self.is_null():
            return None
        else:
            try:
                return self._private_key
            except:
                pass

            raise PermissionError(
                "You do not have permission to see the private key of "
                "this cluster")

    def rotate_keys(self):
        """Call this function to rotate the cluster keys. This is worth
           performing periodically to ensure that the system remains
           secure - this will also rotate the cluster secret
        """
        if self.is_null():
            return

        if Cluster._is_running_service():
            raise PermissionError("Only the client Cluster can rotate keys")

        passphrase = self.passphrase("set_cluster")

        from Acquire.Crypto import PrivateKey as _PrivateKey
        self._oldkeys.append(self._private_key)

        self._private_key = _PrivateKey()
        self._public_key = self._private_key.public_key()
        self._secret = _PrivateKey.random_passphrase()

        args = {"passphrase": passphrase,
                "cluster": self.to_data()}

        self.compute_service().call_function(function="set_cluster",
                                             args=args)

    def encrypt_data(self, data):
        """Encrypt the passed data so that only the daemon running on
           the cluster can decrypt it. This will be encrypted as;

           data = {"cluster_uid" : "CLUSTER_UID",
                   "fingerprint" : "KEY_FINGERPRINT",
                   "encrypted_data" : "ENCRYPTED_DATA"}
        """
        if self.is_null():
            raise PermissionError("You cannot encrypt using a null cluster!")

        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        import json as _json
        return {"cluster_uid": str(self.uid()),
                "fingerprint": str(self.public_key().fingerprint()),
                "encrypted_data": _bytes_to_string(
                                 self.public_key().encrypt(_json.dumps(data)))
                }

    def decrypt_data(self, data):
        """Decrypt the pased data - this will only work in the daemon
           running on the cluster itself. The data must have been
           encrypted using the 'encrypt_data' function of this object.
        """
        if self.is_null():
            raise PermissionError("You cannot decrypt using a null cluster!")

        import json as _json

        if isinstance(data, str):
            data = _json.loads(data)

        try:
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
            cluster_uid = data["cluster_uid"]
            fingerprint = data["fingerprint"]
            data = _string_to_bytes(data["encrypted_data"])
        except Exception as e:
            from Acquire.Crypto import DecryptionError
            raise DecryptionError(
                "The encrypted data is not of the correct format: %s" % str(e))

        if cluster_uid != self.uid():
            from Acquire.Crypto import DecryptionError
            raise DecryptionError(
                "Cannot decrypt the data as it wasn't encrypted for this "
                "cluster - unmatched cluster UID: %s versus %s" %
                (cluster_uid, self.uid()))

        key = self.private_key()

        try:
            i = len(self._oldkeys)
        except:
            i = 0

        while fingerprint != key.fingerprint():
            i = i-1
            if i < 0:
                from Acquire.Crypto import DecryptionError
                raise DecryptionError(
                    "Cannot decrypt the data as we don't recognise the "
                    "fingerprint of the encryption key: %s" % fingerprint)

            key = self._oldkeys[i]

        data = key.decrypt(data)
        return _json.loads(data)

    def get_job(self, uid, start_state="pending", end_state=None,
                passphrase=None):
        """Return the job with specified 'uid' in the specified
           state (start_state) - this will move the job to
           'end_state' if this is specified. If you are on the
           service you need to supply a valid passphrase
        """
        if end_state is None:
            resource = "get_job %s %s" % (uid, start_state)
        else:
            resource = "get_job %s %s->%s" % (uid, start_state, end_state)

        if Cluster._is_running_service():
            self.verify_passphrase(resource=resource, passphrase=passphrase)

            start_state = JobState(start_state)

            if end_state is not None:
                end_state = JobState(end_state)

            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket

            bucket = _get_service_account_bucket()
            key = "compute/%s/%s" % (start_state.value, uid)

            if (end_state is None) or (end_state == start_state):
                try:
                    data = _ObjectStore.get_object_from_json(bucket=bucket,
                                                             key=key)
                except:
                    data = None
            else:
                from Acquire.ObjectStore import get_datetime_now_to_string \
                    as _get_datetime_now_to_string

                try:
                    data = _ObjectStore.take_object_from_json(bucket=bucket,
                                                              key=key)
                    data[end_state.value] = _get_datetime_now_to_string()
                    key = "compute/%s/%s" % (end_state.value, uid)
                    _ObjectStore.set_object_from_json(bucket=bucket, key=key,
                                                      data=data)
                except:
                    data = None

            if data is None:
                raise KeyError(
                    "There is no job with UID %s in state %s" %
                    (uid, start_state.value))

            # the data is a dictionary of the submission time and the
            # job UID
            if uid != data["uid"]:
                raise ValueError("The job info for UID %s is corrupt? %s" %
                                 (uid, data))

            # now load the actual job info
            from Acquire.Compute import ComputeJob as _ComputeJob
            return _ComputeJob.load(uid=uid)
        else:
            passphrase = self.passphrase(resource)
            args = {"uid": str(uid),
                    "passphrase": passphrase,
                    "start_state": str(start_state)}

            if end_state is not None:
                args["end_state"] = str(end_state)

            result = self.compute_service().call_function(function="get_job",
                                                          args=args)

            from Acquire.Compute import ComputeJob as _ComputeJob
            return _ComputeJob.from_data(self.decrypt_data(result["job"]))

    def submit_job(self, uid):
        """Submit the job with specified UID to this cluster.

           On the service this will put the UID of the job into the
           "pending" pool, and will signal the cluster to pull that job

           On the client this will pull the job with that UID from the
           pending pool, moving it to the "submitting" pool and will
           pass this job to the cluster submission system
        """
        if Cluster._is_running_service():
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            from Acquire.ObjectStore import get_datetime_now_to_string \
                as _get_datetime_now_to_string

            bucket = _get_service_account_bucket()
            key = "compute/pending/%s" % uid
            resource = {"pending": _get_datetime_now_to_string(),
                        "uid": uid}

            _ObjectStore.set_object_from_json(bucket, key, resource)
        else:
            # fetch the pending job and change the status to "submitting"
            return self.get_job(uid=uid, start_state="pending",
                                end_state="submitting")

    def get_pending_job_uids(self, passphrase=None):
        """Return the UIDs of all of the jobs that need to be submitted"""
        if self.is_null():
            return []

        if Cluster._is_running_service():
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket

            self.verify_passphrase(resource="get_pending_job_uids",
                                   passphrase=passphrase)

            bucket = _get_service_account_bucket()
            prefix = "compute/pending/"

            uids = _ObjectStore.get_all_object_names(bucket=bucket,
                                                     prefix=prefix,
                                                     without_prefix=True)

            return uids
        else:
            passphrase = self.passphrase(resource="get_pending_job_uids")
            args = {"passphrase": passphrase}
            result = self.compute_service().call_function(
                            function="get_pending_job_uids", args=args)

            return self.decrypt_data(result["job_uids"])

    def to_data(self, passphrase=None):
        """Return a json-serialisable dictionary of this cluster"""
        if self.is_null():
            return {}

        data = {}

        data["uid"] = str(self._uid)
        data["public_key"] = self._public_key.to_data()
        data["secret"] = str(self._secret)

        if passphrase is not None:
            try:
                data["private_key"] = self._private_key.to_data(
                                                    passphrase=passphrase)
            except:
                pass

            try:
                oldkeys = []
                for key in self._oldkeys:
                    oldkeys.append(key.to_data(passphrase=passphrase))

                data["oldkeys"] = oldkeys
            except:
                pass

        return data

    @staticmethod
    def from_data(data, passphrase=None):
        """Return a cluster from the passed json-deserialised data"""
        if data is None or len(data) == 0:
            return Cluster()

        from Acquire.Client import PrivateKey as _PrivateKey
        from Acquire.Client import PublicKey as _PublicKey

        cluster = Cluster()
        cluster._uid = str(data["uid"])
        cluster._public_key = _PublicKey.from_data(data["public_key"])
        cluster._secret = str(data["secret"])

        if passphrase is not None:
            try:
                cluster._private_key = _PrivateKey.from_data(
                                                    data["private_key"],
                                                    passphrase=passphrase)
            except:
                pass

            if "oldkeys" in data:
                try:
                    oldkeys = data["oldkeys"]
                    cluster._oldkeys = []

                    for key in oldkeys:
                        key = _PrivateKey.from_data(key, passphrase=passphrase)
                        cluster._oldkeys.append(key)
                except:
                    pass

        return cluster



#!/bin/env python3


def error_args():
    print("ERROR")


def aq_ls():
    print("ls")


def main():
    import argparse
    import sys

    if len(sys.argv) < 2:
        error_args()

    elif sys.argv[1] == "ls":
        aq_ls()

    else:
        error_args()




__all__ = ["File"]


class File:
    """This class provides a handle to a user's file on a Drive.
       Files exist on Drives, and are read-only on the Drive.
       You update a File by uploading a new version
    """
    def __init__(self):
        self._metadata = None

    def is_null(self):
        """Return whether or not this File is null"""
        return self._metadata is None

    @staticmethod
    def open(metadata, creds=None):
        """Open and return the File from the passed FileMeta. The
           file either needs to be opened by the User with
           specified storage service, or by passing in a
           valid PAR and secret
        """
        from Acquire.Client import FileMeta as _FileMeta

        if not isinstance(metadata, _FileMeta):
            raise TypeError("The metadata must be type FileMeta")

        if creds is not None:
            from Acquire.Client import StorageCreds as _StorageCreds
            if not isinstance(creds, _StorageCreds):
                raise TypeError("The creds must be type StorageCreds")

        from copy import copy as _copy
        metadata = _copy(metadata)

        if creds is None:
            creds = metadata._creds
        else:
            metadata._creds = creds

        if creds.is_par():
            par = creds.par()
            loc = par.location()
            if not loc.is_drive():
                # we need to create a dummy Drive metadata for this file
                from Acquire.Client import DriveMeta as _DriveMeta
                d = _DriveMeta(name="par:%s" % par.uid(),
                               uid=loc.drive_uid())
                metadata._set_drive_metadata(d, creds)

        f = File()
        f._metadata = metadata
        f._creds = creds

        return f

    def metadata(self):
        """Return the metadata for this file"""
        if self.is_null():
            return None
        else:
            from copy import copy as _copy
            return _copy(self._metadata)

    def credentials(self):
        """Return the credentials used to open this file"""
        if self.is_null():
            return None
        else:
            return self._creds

    def __str__(self):
        if self.is_null():
            return "File::null"
        else:
            return "File(name='%s')" % self._metadata.name()

    def chunk_upload(self, aclrules=None):
        """Start a chunk-upload of a new version of this file. This
           will return a chunk-uploader that can be used to upload
           a file chunk-by-chunk
        """
        if self.is_null():
            raise PermissionError("Cannot download a null File!")

        if self._creds is None:
            raise PermissionError("We have not properly opened the file!")

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.ObjectStore import OSPar as _OSPar
        from Acquire.Client import FileMeta as _FileMeta

        uploaded_name = self._metadata.filename()
        drive_uid = self._metadata.drive().uid()

        args = {"filename": uploaded_name,
                "drive_uid": drive_uid}

        if aclrules is not None:
            args["aclrules"] = aclrules.to_data()

        if self._creds.is_user():
            authorisation = _Authorisation(
                        resource="chunk_upload %s" % uploaded_name,
                        user=self._creds.user())

            args["authorisation"] = authorisation.to_data()
        elif self._creds.is_par():
            par = self._creds.par()
            par.assert_valid()
            args["par_uid"] = par.uid()
            args["secret"] = self._creds.secret()
        else:
            raise PermissionError(
                "Either a logged-in user or valid PAR must be provided!")

        if self._creds.is_user():
            privkey = self._creds.user().session_key()
        else:
            from Acquire.Crypto import get_private_key \
                as _get_private_key
            privkey = _get_private_key("parkey")

        args["encryption_key"] = privkey.public_key().to_data()

        # will eventually need to authorise payment...
        storage_service = self._creds.storage_service()

        response = storage_service.call_function(
                                function="open_uploader", args=args)

        filemeta = _FileMeta.from_data(response["filemeta"])

        filemeta._set_drive_metadata(self._metadata._drive_metadata,
                                     self._creds)

        self._metadata = filemeta

        from Acquire.Client import ChunkUploader as _ChunkUploader
        return _ChunkUploader.from_data(response["uploader"],
                                        privkey=privkey,
                                        service=storage_service)

    def upload(self, filename, force_par=False, aclrules=None):
        """Upload 'filename' as the new version of this file"""
        if self.is_null():
            raise PermissionError("Cannot download a null File!")

        if self._creds is None:
            raise PermissionError("We have not properly opened the file!")

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.ObjectStore import OSPar as _OSPar
        from Acquire.Client import FileMeta as _FileMeta
        from Acquire.Storage import FileHandle as _FileHandle

        local_cutoff = None

        if force_par:
            # only upload using a OSPar
            local_cutoff = 0

        uploaded_name = self._metadata.filename()
        drive_uid = self._metadata.drive().uid()

        filehandle = _FileHandle(filename=filename,
                                 remote_filename=uploaded_name,
                                 drive_uid=drive_uid,
                                 aclrules=aclrules,
                                 local_cutoff=local_cutoff)

        try:
            args = {"filehandle": filehandle.to_data()}

            if self._creds.is_user():
                authorisation = _Authorisation(
                            resource="upload %s" % filehandle.fingerprint(),
                            user=self._creds.user())

                args["authorisation"] = authorisation.to_data()
            elif self._creds.is_par():
                par = self._creds.par()
                par.assert_valid()
                args["par_uid"] = par.uid()
                args["secret"] = self._creds.secret()
            else:
                raise PermissionError(
                    "Either a logged-in user or valid PAR must be provided!")

            if not filehandle.is_localdata():
                # we will need to upload against a OSPar, so need to tell
                # the service how to encrypt the OSPar...
                if self._creds.is_user():
                    privkey = self._creds.user().session_key()
                else:
                    from Acquire.Crypto import get_private_key \
                        as _get_private_key
                    privkey = _get_private_key("parkey")

                args["encryption_key"] = privkey.public_key().to_data()

            # will eventually need to authorise payment...
            storage_service = self._creds.storage_service()

            response = storage_service.call_function(
                                    function="upload", args=args)

            filemeta = _FileMeta.from_data(response["filemeta"])

            # if this was a large file, then we will receive a OSPar back
            # which must be used to upload the file
            if not filehandle.is_localdata():
                par = _OSPar.from_data(response["upload_par"])
                par.write(privkey).set_object_from_file(
                                        filehandle.local_filename())
                par.close(privkey)

            filemeta._set_drive_metadata(self._metadata._drive_metadata,
                                         self._creds)

            return filemeta
        except:
            # ensure that any temporary files are removed
            filehandle.__del__()
            raise

    def chunk_download(self, filename=None, version=None,
                       dir=None):
        """Return a ChunkDownloader to download this file
           chunk-by-chunk
        """
        if self.is_null():
            raise PermissionError("Cannot download a null File!")

        if self._creds is None:
            raise PermissionError("We have not properly opened the file!")

        if filename is None:
            filename = self._metadata.name()

        drive_uid = self._metadata.drive().uid()

        if self._creds.is_user():
            privkey = self._creds.user().session_key()
        else:
            from Acquire.Crypto import get_private_key as _get_private_key
            privkey = _get_private_key("parkey")

        args = {"drive_uid": drive_uid,
                "filename": self._metadata.name(),
                "encryption_key": privkey.public_key().to_data(),
                "must_chunk": True}

        if self._creds.is_user():
            from Acquire.Client import Authorisation as _Authorisation
            authorisation = _Authorisation(
                        resource="download %s %s" % (drive_uid,
                                                     self._metadata.name()),
                        user=self._creds.user())
            args["authorisation"] = authorisation.to_data()
        elif self._creds.is_par():
            par = self._creds.par()
            par.assert_valid()
            args["par_uid"] = par.uid()
            args["secret"] = self._creds.secret()

        storage_service = self._creds.storage_service()

        response = storage_service.call_function(
                                function="download", args=args)

        from Acquire.Client import ChunkDownloader as _ChunkDownloader
        downloader = _ChunkDownloader.from_data(response["downloader"],
                                                privkey=privkey,
                                                service=storage_service)

        downloader._start_download(filename=filename, dir=dir)

        return downloader

    def download(self, filename=None, version=None,
                 dir=None, force_par=False):
        """Download this file into the local directory
           the local directory, or 'dir' if specified,
           calling the file 'filename' (or whatever it is called
           on the Drive if not specified). If a local
           file exists with this name, then a new, unique filename
           will be used. This returns the local filename of the
           downloaded file (with full absolute path)

           Note that this only downloads files for which you
           have read-access. If the file is not readable then
           an exception is raised and nothing is returned

           If 'version' is specified then download a specific version
           of the file. Otherwise download the version associated
           with this file object
        """
        if self.is_null():
            raise PermissionError("Cannot download a null File!")

        if self._creds is None:
            raise PermissionError("We have not properly opened the file!")

        if filename is None:
            filename = self._metadata.name()

        drive_uid = self._metadata.drive().uid()

        from Acquire.Client import create_new_file as \
            _create_new_file

        if self._creds.is_user():
            privkey = self._creds.user().session_key()
        else:
            from Acquire.Crypto import get_private_key as _get_private_key
            privkey = _get_private_key("parkey")

        args = {"drive_uid": drive_uid,
                "filename": self._metadata.name(),
                "encryption_key": privkey.public_key().to_data()}

        if self._creds.is_user():
            from Acquire.Client import Authorisation as _Authorisation
            authorisation = _Authorisation(
                        resource="download %s %s" % (drive_uid,
                                                     self._metadata.name()),
                        user=self._creds.user())
            args["authorisation"] = authorisation.to_data()
        elif self._creds.is_par():
            par = self._creds.par()
            par.assert_valid()
            args["par_uid"] = par.uid()
            args["secret"] = self._creds.secret()

        if force_par:
            args["force_par"] = True

        if version is not None:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            args["version"] = _datetime_to_string(version)
        elif self._metadata.version() is not None:
            args["version"] = self._metadata.version()

        storage_service = self._creds.storage_service()

        response = storage_service.call_function(
                                function="download", args=args)

        from Acquire.Client import FileMeta as _FileMeta
        filemeta = _FileMeta.from_data(response["filemeta"])

        if "filedata" in response:
            # we have already downloaded the file to 'filedata'
            filedata = response["filedata"]

            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes
            filedata = _string_to_bytes(response["filedata"])
            del response["filedata"]

            # validate that the size and checksum are correct
            filemeta.assert_correct_data(filedata)

            if filemeta.is_compressed():
                # uncompress the data
                from Acquire.Client import uncompress as _uncompress
                filedata = _uncompress(
                                inputdata=filedata,
                                compression_type=filemeta.compression_type())

            # write the data to the specified local file...
            filename = _create_new_file(filename=filename, dir=dir)
            with open(filename, "wb") as FILE:
                FILE.write(filedata)
                FILE.flush()
        elif "download_par" in response:
            from Acquire.ObjectStore import OSPar as _OSPar
            filename = _create_new_file(filename=filename, dir=dir)
            par = _OSPar.from_data(response["download_par"])
            par.read(privkey).get_object_as_file(filename)
            par.close(privkey)

            # validate that the size and checksum are correct
            filemeta.assert_correct_data(filename=filename)

            # uncompress the file if desired
            if filemeta.is_compressed():
                from Acquire.Client import uncompress as _uncompress
                _uncompress(inputfile=filename,
                            outputfile=filename,
                            compression_type=filemeta.compression_type())

        elif "downloader" in response:
            from Acquire.Client import ChunkDownloader as _ChunkDownloader
            downloader = _ChunkDownloader.from_data(response["downloader"],
                                                    privkey=privkey,
                                                    service=storage_service)

            filename = downloader.download(filename=filename, dir=dir)

        filemeta._copy_credentials(self._metadata)
        self._metadata = filemeta

        return filename

    def list_versions(self, include_metadata=False):
        """Return a list of all of the versions of this file.
           If 'include_metadata' is True then this will include
           the full metadata of every version
        """
        if self.is_null():
            return []

        if self._creds is None:
            raise PermissionError(
                "Cannot list versions of this file as it has not "
                "been properly opened")

        drive_uid = self._metadata.drive().uid()

        if include_metadata:
            include_metadata = True
        else:
            include_metadata = False

        filename = self._metadata.name()

        args = {"drive_uid": drive_uid,
                "include_metadata": include_metadata,
                "filename": filename}

        if self._creds.is_user():
            from Acquire.Client import Authorisation as _Authorisation
            authorisation = _Authorisation(
                                    resource="list_versions %s" % filename,
                                    user=self._creds.user())
            args["authorisation"] = authorisation.to_data()
        elif self._creds.is_par():
            par = self._creds.par()
            par.assert_valid()
            args["par_uid"] = par.uid()
            args["secret"] = self._creds.secret()

        storage_service = self._creds.storage_service()

        response = storage_service.call_function(function="list_versions",
                                                 args=args)

        from Acquire.ObjectStore import string_to_list \
            as _string_to_list
        from Acquire.Storage import FileMeta as _FileMeta

        versions = _string_to_list(response["versions"], _FileMeta)

        for version in versions:
            version._copy_credentials(self._metadata)

        return versions




__all__ = ["DirMeta"]


class DirMeta:
    """This is a lightweight class that holds the metadata about
       a particular directory in a Drive. Note that directories
       are not versioned, and they do not contain any ACLs
       (the ACLs are attached to files or drives). Directories
       just provide a convenient way to group together a set
       of files in a Drive
    """
    def __init__(self, name=None):
        """Construct, specifying the name of the directory"""
        self._name = name
        self._drive_metadata = None
        self._creds = None

    def __str__(self):
        """Return a string representation"""
        if self.is_null():
            return "DirMeta::null"
        else:
            return "DirMeta(%s)" % self._name

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        """Comparison equals"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def open(self, creds=None):
        """Open and return the File associated with this metadata"""
        from Acquire.Client import Directory as _Directory
        return _Directory.open(metadata=self, creds=creds)

    def is_null(self):
        """Return whether or not this is null"""
        return self._name is None

    def _copy_credentials(self, other):
        """Copy the drive metadata and credentials from 'other' to self"""
        if not isinstance(other, DirMeta):
            raise TypeError("other should be a DirMeta")

        self._drive_metadata = other._drive_metadata
        self._creds = other._creds

    def _set_drive_metadata(self, metadata, creds=None):
        """Internal function called by "Drive" to store the metadata
           of the drive that contains the Directory described by this metadata
        """
        from Acquire.Client import DriveMeta as _DriveMeta
        if not isinstance(metadata, _DriveMeta):
            raise TypeError("The metadata must be type DriveMeta")

        if creds is None:
            creds = metadata._creds
        else:
            if creds is not metadata._creds:
                from copy import copy as _copy
                metadata = _copy(metadata)
                metadata._creds = creds

        self._drive_metadata = metadata
        self._creds = metadata._creds

    def location(self):
        """Return a global location for this directory. This is unique
           for this directory and can be used to locate this directory from
           any other service.
        """
        if self.is_null():
            return None
        elif self._drive_metadata is None:
            raise PermissionError(
                "Cannot generate the location as we don't know "
                "which drive this file has come from!")

        from Acquire.Client import Location as _Location
        return _Location(drive_guid=self._drive_metadata.guid(),
                         filename=self.name())

    def name(self):
        """Return the name of the directory"""
        return self._name

    def drive(self):
        """Return the metadata for the drive that contains this file"""
        if self.is_null():
            return None
        elif self._drive_metadata is None:
            raise PermissionError(
                "Cannot get the drive metadata as we don't know "
                "which drive this directory has come from!")
        else:
            from copy import copy as _copy
            return _copy(self._drive_metadata)

    def to_data(self):
        """Return a json-serialisable dictionary of this object"""
        data = {}

        if self.is_null():
            return data

        data["name"] = str(self._name)

        return data

    @staticmethod
    def from_data(data):
        """Return a new DirMeta constructed from the passed json-deserialised
           dictionary
        """
        d = DirMeta()

        if data is not None and len(data) > 0:
            d._name = data["name"]

        return d




__all__ = ["UserAccount"]

_user_root = "identity/users"


def _encode_username(username):
    """This function returns an encoded (sanitised) version of
        the username. This will ensure that the username
        is valid (must be between 3 and 50 characters).
        The sanitised username is the encoded version,
        meaning that a user can use a unicode (emoji)
        username if they so desire
    """
    if username is None:
        return None

    if len(username) < 3 or len(username) > 150:
        from Acquire.Identity import UsernameError
        raise UsernameError(
            "The username must be between 3 and 150 characters!")

    from Acquire.ObjectStore import string_to_encoded \
        as _string_to_encoded

    return _string_to_encoded(username)


class UserAccount:
    """This class holds all information about a user's account,
       e.g. their username, the sanitised username for the person
       on the system, their account keys, status etc.

       This data can be serialised to an from json to allow
       easy saving a retrieval from an object store
    """

    def __init__(self, username=None, user_uid=None,
                 private_key=None, status=None):
        """Construct from the passed username"""
        self._username = username
        self._uid = user_uid
        self._privkey = private_key
        self._status = status

    @staticmethod
    def create(username, password, _service_uid=None,
               _service_public_key=None):
        """Create a new account with username 'username', which will
           be secured using the passed password.

           Note that this will create an account with a specified
           user UID, meaning that different users can have the same
           username. We identify the right user via the combination
           of username, password and OTP code.

           Normally the UID of the service, and the skeleton key
           used to encrypt the backup password are obtained
           directly from the service. However, when initialising
           a new service we must pass these directly. In those
           cases, pass the object using _service_uid and
           _service_public_key

           This returns a tuple of the user_uid and OTP for the
           newly-created account
        """
        from Acquire.ObjectStore import create_uuid as _create_uuid
        from Acquire.Crypto import PrivateKey as _PrivateKey
        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        from Acquire.Identity import UserCredentials as _UserCredentials
        from Acquire.ObjectStore import get_datetime_now_to_string \
            as _get_datetime_now_to_string

        if _service_public_key is None:
            from Acquire.Service import get_this_service as _get_this_service
            service_pubkey = _get_this_service().public_skeleton_key()
            assert(service_pubkey is not None)
        else:
            service_pubkey = _service_public_key

        if not isinstance(service_pubkey, _PublicKey):
            raise TypeError("The service public key must be type PublicKey")

        if _service_uid is None:
            from Acquire.Service import get_this_service \
                as _get_this_service
            service_uid = _get_this_service(need_private_access=False).uid()
        else:
            service_uid = _service_uid

        # create a UID for this new user
        user_uid = _create_uuid()

        # now create the primary password for this user and use
        # this to encrypt the special keys for this user
        privkey = _PrivateKey(name="user_secret_key %s %s" % (username,
                                                              user_uid))
        primary_password = _PrivateKey.random_passphrase()

        bucket = _get_service_account_bucket()

        # now create the credentials used to validate a login
        otp = _UserCredentials.create(user_uid=user_uid,
                                      password=password,
                                      primary_password=primary_password)

        # create the user account
        user = UserAccount(username=username, user_uid=user_uid,
                           private_key=privkey, status="active")

        # now save a lookup from the username to this user_uid
        # (many users can have the same username). Use this lookup
        # to hold a recovery password for this account
        recovery_password = _bytes_to_string(
                                service_pubkey.encrypt(primary_password))

        key = "%s/names/%s/%s" % (_user_root, user.encoded_name(), user_uid)
        _ObjectStore.set_string_object(bucket=bucket, key=key,
                                       string_data=recovery_password)

        # now save a lookup from the hashed username+password
        # to the user_uid, so that we can
        # quickly find matching user_uids (expect few people will have
        # exactly the same username and password). This will
        # save the exact time this username-password combination
        # was set
        encoded_password = _UserCredentials.hash(username=username,
                                                 password=password,
                                                 service_uid=service_uid)

        key = "%s/passwords/%s/%s" % (_user_root, encoded_password, user_uid)
        _ObjectStore.set_string_object(
                                bucket=bucket, key=key,
                                string_data=_get_datetime_now_to_string())

        # finally(!) save the account itself to the object store
        key = "%s/uids/%s" % (_user_root, user_uid)
        data = user.to_data(passphrase=primary_password)
        _ObjectStore.set_object_from_json(bucket=bucket,
                                          key=key,
                                          data=data)

        # return the OTP and user_uid
        return (user_uid, otp)

    @staticmethod
    def login(credentials, user_uid=None, remember_device=False):
        """Login to the session with specified 'short_uid' with the
           user with passed 'username' and 'credentials',
           optionally specifying the user_uid
        """
        if user_uid is None:
            # find all of the user_uids of accounts with this
            # username+password combination
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            from Acquire.Client import Credentials as _Credentials
            from Acquire.Identity import UserCredentials as _UserCredentials
            from Acquire.Service import get_this_service as _get_this_service

            if not isinstance(credentials, _Credentials):
                raise TypeError("The credentials must be type Credentials")

            bucket = _get_service_account_bucket()

            encoded_password = _UserCredentials.hash(
                                        username=credentials.username(),
                                        password=credentials.password())

            prefix = "%s/passwords/%s/" % (_user_root, encoded_password)

            try:
                names = _ObjectStore.get_all_object_names(bucket=bucket,
                                                          prefix=prefix)
            except:
                names = []

            user_uids = []
            for name in names:
                user_uids.append(name.split("/")[-1])
        else:
            user_uids = [user_uid]

        if len(user_uids) == 0:
            from Acquire.Identity import UserValidationError
            raise UserValidationError("No user with name '%s'" %
                                      credentials.username())

        from Acquire.Identity import UserCredentials as _UserCredentials
        return _UserCredentials.login(credentials=credentials,
                                      user_uids=user_uids,
                                      remember_device=remember_device)

    def __str__(self):
        return "UserAccount(name : %s)" % self._username

    def name(self):
        """Return the name of this account"""
        return self._username

    def username(self):
        """Synonym for 'name'"""
        return self.name()

    def encoded_name(self):
        """Return the encoded (sanitised) username"""
        return _encode_username(self._username)

    def uid(self):
        """Return the globally unique ID for this account"""
        return self._uid

    def login_root_url(self):
        """Return the root URL used to log into this account"""
        from Acquire.Service import get_this_service as _get_this_service
        return _get_this_service().canonical_url()

    def is_valid(self):
        """Return whether or not this is a valid account"""
        return not (self._status is None)

    def is_active(self):
        """Return whether or not this is an active account"""
        if self._status is None:
            return False
        else:
            return self._status == "active"

    def public_key(self):
        """Return the lines of the public key for this account"""
        return self._privkey.public_key()

    def private_key(self):
        """Return the lines of the private key for this account"""
        return self._privkey

    def status(self):
        """Return the status for this account"""
        if self._status is None:
            return "invalid"

        return self._status

    def to_data(self, passphrase, mangleFunction=None):
        """Return a data representation of this object (dictionary)"""
        if self._username is None:
            return None

        data = {}
        data["username"] = self._username
        data["status"] = self._status
        data["uid"] = self._uid
        data["private_key"] = self._privkey.to_data(
                                        passphrase=passphrase,
                                        mangleFunction=mangleFunction)

        return data

    @staticmethod
    def from_data(data, passphrase, mangleFunction=None):
        """Return a UserAccount constructed from the passed
           data (dictionary)
        """

        user = UserAccount()

        if data is not None and len(data) > 0:
            from Acquire.Crypto import PrivateKey as _PrivateKey

            user._username = data["username"]
            user._status = data["status"]
            user._uid = data["uid"]
            user._privkey = _PrivateKey.from_data(
                                            data=data["private_key"],
                                            passphrase=passphrase,
                                            mangleFunction=mangleFunction)

        return user




__all__ = ["ChunkUploader"]


class ChunkUploader:
    """This class is used to control the chunked uploading
       of a file. This allows a file to be uploaded
       chunk by chunk (bit by bit). This is useful, e.g.
       to upload a file as it is being written
    """
    def __init__(self, drive_uid=None, file_uid=None):
        """Create a new ChunkUploader that uploads the specified
           file to the specified drive
        """
        self._drive_uid = None
        self._file_uid = None
        self._chunk_idx = None
        self._service = None

        if drive_uid is not None:
            self._drive_uid = str(drive_uid)
            self._file_uid = str(file_uid)
            from Acquire.Crypto import PrivateKey as _PrivateKey
            self._secret = _PrivateKey.random_passphrase()

    def __exit__(self, exception_type, exception_value, traceback):
        """Ensure that we close the file"""
        if self.is_open():
            self.close()

    def __del__(self):
        """Make sure that we close the file before deleting this object"""
        if self.is_open():
            self.close()

    def is_null(self):
        """Return whether or not this is null"""
        return self._drive_uid is None or self._file_uid is None

    def secret(self):
        """Return the secret used to authenticate the upload"""
        return self._secret

    def service(self):
        """Return the service that created this uploader"""
        return self._service

    def upload(self, chunk):
        """Upload the next chunk of the file"""
        if self.is_null():
            raise PermissionError("Cannot upload a chunk to a null uploader!")

        service = self.service()

        if service is None:
            raise PermissionError("Cannot upload a chunk to a null service!")

        # first, compress the chunk
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        from Acquire.Crypto import Hash as _Hash
        import bz2 as _bz2

        if isinstance(chunk, str):
            chunk = chunk.encode("utf-8")

        chunk = _bz2.compress(chunk)
        md5 = _Hash.md5(chunk)
        chunk = _bytes_to_string(chunk)

        if self._chunk_idx is None:
            self._chunk_idx = 0
        else:
            self._chunk_idx = self._chunk_idx + 1

        secret = _Hash.multi_md5(self._secret,
                                 "%s%s%d" % (self._drive_uid,
                                             self._file_uid,
                                             self._chunk_idx))

        args = {}
        args["drive_uid"] = self._drive_uid
        args["file_uid"] = self._file_uid
        args["chunk_index"] = self._chunk_idx
        args["secret"] = secret
        args["data"] = chunk
        args["checksum"] = md5

        service.call_function(function="upload_chunk", args=args)

    def is_open(self):
        """Return whether or not the file is open (has been written to)"""
        return self._chunk_idx is not None

    def close(self):
        """Close the uploader - this will finalise the file"""
        if self.is_open():
            args = {"drive_uid": self._drive_uid,
                    "file_uid": self._file_uid,
                    "secret": self._secret}

            self.service().call_function(function="close_uploader",
                                         args=args)

            self._chunk_idx = None
            self._secret = None
            self._drive_uid = None
            self._file_uid = None

    def to_data(self, pubkey=None):
        """Return a json-serialisable dictionary of the object. If
           'pubkey' is supplied, then sensitive data used by the
           uploader is encrypted
        """
        if self.is_null():
            return {}

        data = {}

        data["drive_uid"] = self._drive_uid
        data["file_uid"] = self._file_uid
        data["secret"] = self._secret

        if pubkey is not None:
            from Acquire.Crypto import PublicKey as _PublicKey
            from Acquire.ObjectStore import bytes_to_string \
                as _bytes_to_string
            import json as _json

            if not isinstance(pubkey, _PublicKey):
                raise TypeError("pubkey must be type PublicKey")

            d = _bytes_to_string(pubkey.encrypt(_json.dumps(data)))

            data = {}
            data["is_encrypted"] = True
            data["data"] = d

        return data

    @staticmethod
    def from_data(data, privkey=None, service=None):
        """Return a ChunkUploader from a json-deserialised dictionary.
           If this was encrypted then you need to supply a private
           key to decrypt the sensitive data
        """
        if data is None or len(data) == 0:
            return ChunkUploader()

        c = ChunkUploader()

        try:
            is_encrypted = data["is_encrypted"]
        except:
            is_encrypted = False

        if is_encrypted:
            if privkey is None:
                raise PermissionError("Cannot decode as encrypted!")

            from Acquire.Crypto import PrivateKey as _PrivateKey
            if not isinstance(privkey, _PrivateKey):
                raise TypeError("privkey must be type PrivateKey")

            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes
            import json as _json
            data = _json.loads(privkey.decrypt(_string_to_bytes(data["data"])))

        c._drive_uid = data["drive_uid"]
        c._file_uid = data["file_uid"]
        c._secret = data["secret"]
        c._service = service

        return c




__all__ = ["Credentials"]


class Credentials:
    """This class holds user credentials that are sent to the
       service to authenticate users
    """
    def __init__(self, username=None,
                 short_uid=None, device_uid=None,
                 password=None, otpcode=None):
        """Construct to contain the passed credentials"""
        self._username = username

        if username is not None:
            self._short_uid = short_uid
            self._device_uid = device_uid
            self._password = password
            self._otpcode = otpcode

    def is_null(self):
        """Return whether or not these credentials are null

           Returns:
                bool: True if null, else False
        """
        return self._username is None

    def username(self):
        """Return the decoded username

           Returns:
                str: Username held for these credentials

        """
        if self.is_null():
            return None
        else:
            return self._username

    def short_uid(self):
        """Return the decoded session short UID

           Returns:
                str: Short UID of session
        """
        if self.is_null():
            return None
        else:
            return self._short_uid

    def device_uid(self):
        """Return the decoded device UID

           Returns:
                str: UID for device
        """
        if self.is_null():
            return None
        else:
            return self._device_uid

    def password(self):
        """Return the decoded password

           Returns:
                str: Decoded password
        """
        if self.is_null():
            return None
        else:
            return self._password

    def otpcode(self):
        """Return the decoded one time password code (otpcode)

           Returns:
                str: OTP code for session
        """
        if self.is_null():
            return None
        else:
            return self._otpcode

    def to_data(self, identity_uid):
        """Package these credentials into a secure package that
           can be encoded to json and sent to the service.
           Note that you must supply the UID of the identity
           service that you will send this package to...

           Args:
                identity_uid (str): UID of identity service
           Returns:
                str: String containing credential data
        """
        if self.is_null():
            return None

        return Credentials.package(identity_uid=identity_uid,
                                   short_uid=self._short_uid,
                                   username=self._username,
                                   password=self._password,
                                   otpcode=self._otpcode,
                                   device_uid=self._device_uid)

    @staticmethod
    def from_data(data, username, short_uid, random_sleep=150):
        """Unpackage the passed data that has been deserialised from
           json and return the credentials. You need to pass in
           the username and short_uid that you expect to see.
           The random_sleep adds a random sleep to disrupt
           timing attacks

           Args:
                data (str): Data to create credentials from
                username (str): Username for credentials
                short_uid (str): Short UID to use
                random_sleep (int, default=150): Integer used
                to generate a random sleep time
           Returns:
                Credentials: Credentials object created from data
        """
        result = Credentials.unpackage(data=data, username=username,
                                       short_uid=short_uid,
                                       random_sleep=random_sleep)

        return Credentials(username=result["username"],
                           short_uid=result["short_uid"],
                           device_uid=result["device_uid"],
                           password=result["password"],
                           otpcode=result["otpcode"])

    def assert_matching_username(self, username):
        """Assert that the passed username matches that stored
           in these credentials

           Args:
                username (str): Username to compare
           Returns:
                None
        """
        if self.is_null() or self._username != username:
            raise PermissionError(
                "Disagreement for the username for the matched "
                "credentials")

    @staticmethod
    def encode_device_uid(encoded_password, device_uid):
        """Simple function that takes an existing encoded password,
           and then additionally encodes this using the device_uid

           Args:
                encoded_password (str): Encoded password
                device_uid (str): UID for device
           Returns:
                str: Password encoded with device UID

        """
        if device_uid is None or encoded_password is None:
            return encoded_password

        from Acquire.Crypto import Hash as _Hash
        result = _Hash.md5(encoded_password + device_uid)
        return result

    @staticmethod
    def encode_password(password, identity_uid, device_uid=None):
        """Simple function that creates an MD5 hash of the password,
           salted using the passed identity_uid and (optionally)
           the device_uid

           Args:
                password (str): Password to hash
                identity_uid (str): UID to use as salt
                device_uid (str, default=None): Device UID to use
                as additional salt
           Returns:
                str: Hashed and salted password
        """
        from Acquire.Crypto import Hash as _Hash

        encoded_password = _Hash.multi_md5(identity_uid, password)

        encoded_password = Credentials.encode_device_uid(
                                        encoded_password=encoded_password,
                                        device_uid=device_uid)

        return encoded_password

    @staticmethod
    def package(identity_uid, short_uid, username, password, otpcode,
                device_uid=None):
        """Package up the passed credentials so that they can be sent
           to a server for verification. We employ the following
           steps to make it harder for someone to steal the user's
           password:

            1. An MD5 of the password ("password") is generated, salted with
               the UID of the identity service ("identity_uid"), and,
               optionally, the UID of this device ("device_uid")

            2. A symmetric key is generated from the combined MD5s
               of the user's login name (username) and the short UID of
               this login session (short_uid). This is used to encrypt
               the MD5's password and one-time password code ("otpcode").
               The username and session UID are not sent to the server,
               so an attacker must know what these are to extract
               this information.

            3. Also remember that all communication with a service is
               encrypted using the service's public key, and tranmission
               of data should also be sent over HTTPS.

           Args:
                identity_uid (str): UID of the identity service
                short_uid (str): UID of the login session
                username (str): Username for this session
                password (str): Password for user
                otpcode (str): OTP code for session
                device_uid (str): UID for device
           Returns:
                str: JSON serialisable string
        """
        from Acquire.Crypto import Hash as _Hash
        from Acquire.Crypto import SymmetricKey as _SymmetricKey
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string

        if username is None or password is None or otpcode is None:
            raise PermissionError(
                "You must supply a username, password and otpcode "
                "to be able to log in!")

        encoded_password = Credentials.encode_password(
                                            identity_uid=identity_uid,
                                            device_uid=device_uid,
                                            password=password)

        # if the device_uid is not set, then create a random one
        # so that an attacker does not know...
        if device_uid is None:
            from Acquire.ObjectStore import create_uuid as _create_uuid
            device_uid = _create_uuid()

        data = [encoded_password, device_uid, otpcode]
        string_data = "|".join(data)

        uname_shortid = _Hash.md5(username) + _Hash.md5(short_uid)

        data = _SymmetricKey(symmetric_key=uname_shortid).encrypt(string_data)
        result = _bytes_to_string(data)
        return result

    @staticmethod
    def unpackage(data, username, short_uid, random_sleep=150):
        """Unpackage the credentials data packaged using "package" above,
           assuming that this data was packaged for the user login
           name "username" and for the session with short UID "short_uid".

           This will return a dictionary containing:

           username: Login name of the user
           short_uid: Short UID of the login session
           device_uid: UID of the login device (this will be random if it
                       was not set by the user)
           password: The MD5 of the password, salted using the UID of the
                     identity service, and optionally the device_uid
           otpcode: The one-time-password code for this login

           To make timing-based attacks harder, you can set 'random_sleep'
           to add an additional random sleep of up to 'random_sleep'
           milliseconds onto the end of the unpackage function

           Args:
                data (str): String of data containing credentials
                username (str): Username for session
                short_uid (str): UID for session
                random_sleep (int, default=150): Integer used to
                generate a random sleep to prevent timing attacks
           Returns:
                dict: Dictionary containing credentials

        """
        from Acquire.Crypto import Hash as _Hash
        from Acquire.Crypto import SymmetricKey as _SymmetricKey
        from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string

        uname_shortid = _Hash.md5(username) + _Hash.md5(short_uid)

        data = _string_to_bytes(data)

        try:
            data = _SymmetricKey(symmetric_key=uname_shortid).decrypt(data)
        except:
            data = None

        if data is None:
            raise PermissionError("Cannot unpackage/decrypt the credentials")

        data = data.split("|")

        if len(data) < 3:
            raise PermissionError("Invalid credentials! %s" % data)

        result = {"username": username,
                  "short_uid": short_uid,
                  "device_uid": data[1],
                  "password": data[0],
                  "otpcode": data[2]}

        if random_sleep is not None:
            import random as _random
            import time as _time
            random_sleep = _random.randint(0, random_sleep)
            _time.sleep(0.001 * random_sleep)

        return result



"""
Acquire: (C) Christopher Woods 2018

This module implements all of the classes and functions necessary to build the
client (user-facing) interfaces for Acquire
"""

from ._qrcode import *
from ._user import *
from ._account import *
from ._drive import *
from ._file import *
from ._fileops import *
from ._chunkuploader import *
from ._chunkdownloader import *
from ._par import *
from ._location import *
from ._wallet import *
from ._errors import *
from ._cheque import *
from ._service import *
from ._credentials import *
from ._storagecreds import *

# The below objects are useful for the client, so are pulled into
# this module to discourage people using the other Acquire modules
# directly... One day I want to lazy-load all of these...
from Acquire.Crypto import PublicKey, PrivateKey, OTP
from Acquire.Identity import Authorisation, ACLRule, ACLRules, ACLUserRules, \
                             ACLGroupRules, ACLRuleOperation
from Acquire.Storage import DirMeta, FileMeta, DriveMeta

try:
    if __IPYTHON__:
        def _set_printer(C):
            """Function to tell ipython to use __str__ if available"""
            get_ipython().display_formatter.formatters['text/plain'].for_type(
                C,
                lambda obj, p, cycle: p.text(str(obj) if not cycle else '...')
                )

        import sys as _sys
        import inspect as _inspect

        _clsmembers = _inspect.getmembers(_sys.modules[__name__],
                                          _inspect.isclass)

        for _clsmember in _clsmembers:
            _set_printer(_clsmember[1])
except:
    pass
